import pathlib
import sys
from pathlib import Path
from typing import Optional

import typer

from robotoff.types import ObjectDetectionModel, PredictionType, WorkerQueue

app = typer.Typer()


@app.command()
def run_scheduler():
    """Launch the scheduler service."""
    from robotoff import scheduler
    from robotoff.utils import get_logger

    # Defining a root logger
    get_logger()
    scheduler.run()


@app.command()
def run_worker(
    queues: list[WorkerQueue] = typer.Argument(
        ..., help="Names of the queues to listen to"
    ),
    burst: bool = typer.Option(
        False, help="Run in burst mode (quit after all work is done)"
    ),
):
    """Launch a worker."""
    from robotoff.workers.main import run

    run(queues=[x.value for x in queues], burst=burst)


@app.command()
def regenerate_ocr_insights(
    barcode: str = typer.Argument(..., help="Barcode of the product")
) -> None:
    """Regenerate OCR predictions/insights for a specific product and import
    them."""
    from robotoff import settings
    from robotoff.insights import importer
    from robotoff.insights.extraction import (
        DEFAULT_OCR_PREDICTION_TYPES,
        extract_ocr_predictions,
    )
    from robotoff.models import db
    from robotoff.off import generate_json_ocr_url
    from robotoff.products import get_product
    from robotoff.utils import get_logger

    logger = get_logger()

    product = get_product(barcode, ["images"])
    if product is None:
        raise ValueError(f"product not found: {barcode}")

    predictions = []
    for image_id in product["images"]:
        if not image_id.isdigit():
            continue

        ocr_url = generate_json_ocr_url(barcode, image_id)
        predictions += extract_ocr_predictions(
            barcode, ocr_url, DEFAULT_OCR_PREDICTION_TYPES
        )

    with db:
        imported = importer.import_insights(predictions, settings.OFF_SERVER_DOMAIN)

    logger.info("Import finished, %s insights imported", imported)


@app.command()
def generate_ocr_predictions(
    input_path: Path = typer.Argument(
        ..., help="Path to a (gzipped-)OCR JSONL archive"
    ),
    prediction_type: PredictionType = typer.Argument(
        ..., help="Type of the predictions to generate (label, brand,...)"
    ),
    output: Optional[Path] = typer.Option(
        None,
        help="File to write output to, stdout if not specified. Gzipped output are supported.",
        dir_okay=False,
        writable=True,
    ),
) -> None:
    """Generate OCR predictions of the requested type."""
    from robotoff.cli import insights
    from robotoff.utils import get_logger

    get_logger()
    insights.run_from_ocr_archive(input_path, prediction_type, output)


@app.command()
def predict_category(output: str) -> None:
    """Predict categories from the product JSONL dataset stored in `datasets`
    directory."""
    from robotoff import settings
    from robotoff.prediction.category.matcher import predict_from_dataset
    from robotoff.products import ProductDataset
    from robotoff.utils import dump_jsonl

    dataset = ProductDataset(settings.JSONL_DATASET_PATH)
    insights = predict_from_dataset(dataset)
    dict_insights = (i.to_dict() for i in insights)
    dump_jsonl(output, dict_insights)


@app.command()
def download_dataset(minify: bool = False) -> None:
    """Download Open Food Facts dataset and save it in `datasets` directory."""
    from robotoff.products import fetch_dataset, has_dataset_changed
    from robotoff.utils import get_logger

    get_logger()

    if has_dataset_changed():
        fetch_dataset(minify)


@app.command()
def categorize(
    barcode: str,
    deepest_only: bool = False,
) -> None:
    """Predict product categories based on the neural category classifier.

    deepest_only: controls whether the returned predictions should only contain the deepmost
    categories for a predicted taxonomy chain.
    For example, if we predict 'fresh vegetables' -> 'legumes' -> 'beans' for a product,
    setting deepest_only=True will return 'beans'."""
    from robotoff.prediction.category.neural.category_classifier import (
        CategoryClassifier,
    )
    from robotoff.products import get_product
    from robotoff.taxonomy import TaxonomyType, get_taxonomy

    product = get_product(barcode)
    if product is None:
        print(f"Product {barcode} not found")
        return

    predictions = CategoryClassifier(get_taxonomy(TaxonomyType.category.name)).predict(
        product, deepest_only
    )

    if predictions:
        for prediction in predictions:
            print(f"{prediction.value_tag}: {prediction.data['confidence']}")
    else:
        print(f"Nothing predicted for product {barcode}")


@app.command()
def import_insights(
    prediction_type: Optional[PredictionType] = typer.Option(
        None,
        help="Type of the prediction to generate, only needed when --generate-from is used",
    ),
    batch_size: int = typer.Option(
        128,
        help="Number of insights that are imported in each atomic SQL transaction",
    ),
    input_path: Optional[pathlib.Path] = typer.Option(
        None,
        help="Input path of the JSONL archive, is incompatible with --generate-from",
    ),
    generate_from: Optional[pathlib.Path] = typer.Option(
        None, help="Input path of the OCR archive, is incompatible with --input-path"
    ),
) -> None:
    """Import insights from a prediction JSONL archive (with --input-path
    option), or generate them on the fly from an OCR archive (with
    --generate-from option)."""
    import tqdm
    from more_itertools import chunked

    from robotoff import settings
    from robotoff.cli.insights import generate_from_ocr_archive, insights_iter
    from robotoff.insights import importer
    from robotoff.models import db
    from robotoff.utils import get_logger

    logger = get_logger()

    if generate_from is not None:
        logger.info(f"Generating and importing insights from {generate_from}")
        if prediction_type is None:
            sys.exit("Required option: --prediction-type")

        predictions = generate_from_ocr_archive(generate_from, prediction_type)
    elif input_path is not None:
        logger.info(f"Importing insights from {input_path}")
        predictions = insights_iter(input_path)
    else:
        raise ValueError("--generate-from or --input-path must be provided")

    imported = 0
    with db.connection_context():
        for prediction_batch in tqdm.tqdm(
            chunked(predictions, batch_size), desc="prediction batch"
        ):
            # Create a new transaction for every batch
            with db.atomic():
                batch_imported = importer.import_insights(
                    prediction_batch,
                    settings.OFF_SERVER_DOMAIN,
                )
                logger.info(f"{batch_imported} insights imported in batch")
                imported += batch_imported

    logger.info(f"{imported} insights imported")


@app.command()
def refresh_insights(
    barcode: Optional[str] = typer.Option(
        None,
        help="Refresh a specific product. If not provided, all products are updated",
    ),
    batch_size: int = typer.Option(
        100, help="Number of products to send in a worker tasks"
    ),
):
    """Refresh insights based on available predictions.

    If a `barcode` is provided, only the insights of this product is
    refreshed, otherwise insights of all products are refreshed.
    """
    import tqdm
    from more_itertools import chunked
    from peewee import fn

    from robotoff import settings
    from robotoff.insights.importer import refresh_insights as refresh_insights_
    from robotoff.models import Prediction as PredictionModel
    from robotoff.models import db
    from robotoff.utils import get_logger
    from robotoff.workers.queues import enqueue_job, low_queue
    from robotoff.workers.tasks import refresh_insights_job

    logger = get_logger()

    if barcode is not None:
        logger.info(f"Refreshing product {barcode}")
        imported = refresh_insights_(barcode, settings.OFF_SERVER_DOMAIN)
        logger.info(f"Refreshed insights: {imported}")
    else:
        logger.info("Launching insight refresh on full database")
        with db:
            barcodes = [
                barcode
                for (barcode,) in PredictionModel.select(
                    fn.Distinct(PredictionModel.barcode)
                ).tuples()
            ]

        batches = list(chunked(barcodes, batch_size))
        confirm = typer.confirm(
            f"{len(batches)} jobs are going to be launched, confirm?"
        )

        if not confirm:
            return

        logger.info("Adding refresh_insights jobs in queue...")
        for barcode_batch in tqdm.tqdm(batches, desc="barcode batch"):
            enqueue_job(
                refresh_insights_job,
                low_queue,
                job_kwargs={"result_ttl": 0, "timeout": "5m"},
                barcodes=barcode_batch,
                server_domain=settings.OFF_SERVER_DOMAIN,
            )


@app.command()
def import_images_in_db(
    batch_size: int = typer.Option(
        500, help="Number of items to send in a worker tasks"
    ),
):
    """Make sure that every image available in MongoDB is saved in `image`
    table."""
    import tqdm
    from more_itertools import chunked

    from robotoff import settings
    from robotoff.models import ImageModel, db
    from robotoff.off import generate_image_path
    from robotoff.products import get_product_store
    from robotoff.utils import get_logger
    from robotoff.workers.queues import enqueue_job, low_queue
    from robotoff.workers.tasks.import_image import save_image_job

    logger = get_logger()

    with db:
        logger.info("Fetching existing images in DB...")
        existing_images = set(
            ImageModel.select(ImageModel.barcode, ImageModel.image_id).tuples()
        )

    store = get_product_store()
    to_add = []
    for product in tqdm.tqdm(
        store.iter_product(projection=["images", "code"]), desc="product"
    ):
        barcode = product.barcode
        for image_id in (id_ for id_ in product.images.keys() if id_.isdigit()):
            if (barcode, image_id) not in existing_images:
                to_add.append((barcode, generate_image_path(barcode, image_id)))

    batches = list(chunked(to_add, batch_size))
    if typer.confirm(
        f"{len(batches)} add image jobs are going to be launched, confirm?"
    ):
        for batch in tqdm.tqdm(batches, desc="job"):
            enqueue_job(
                save_image_job,
                low_queue,
                job_kwargs={"result_ttl": 0},
                batch=batch,
                server_domain=settings.OFF_SERVER_DOMAIN,
            )


@app.command()
def run_object_detection_model(
    model_name: ObjectDetectionModel = typer.Argument(
        ..., help="Name of the object detection model"
    ),
    limit: Optional[int] = typer.Option(None, help="Maximum numbers of job to launch"),
):
    """Launch object detection model jobs on all missing images (images
    without an ImagePrediction item for this model) in DB."""
    from typing import Callable

    import tqdm
    from peewee import JOIN

    from robotoff import settings
    from robotoff.models import ImageModel, ImagePrediction, db
    from robotoff.off import generate_image_url
    from robotoff.workers.queues import enqueue_job, low_queue
    from robotoff.workers.tasks.import_image import (
        run_logo_object_detection,
        run_nutriscore_object_detection,
        run_nutrition_table_object_detection,
    )

    kwargs = {}
    if model_name == ObjectDetectionModel.universal_logo_detector:
        func: Callable = run_logo_object_detection
        kwargs["process_logos"] = False
    elif model_name == ObjectDetectionModel.nutrition_table:
        func = run_nutrition_table_object_detection
    else:
        func = run_nutriscore_object_detection

    with db:
        query = (
            ImageModel.select(ImageModel.barcode, ImageModel.id)
            .join(
                ImagePrediction,
                JOIN.LEFT_OUTER,
                on=(
                    (ImagePrediction.image_id == ImageModel.id)
                    & (ImagePrediction.model_name == model_name)
                ),
            )
            .where(ImagePrediction.model_name.is_null())
            .tuples()
        )
        if limit:
            query = query.limit(limit)
        missing_items = list(query)

    if limit:
        missing_items = missing_items[:limit]

    if typer.confirm(f"{len(missing_items)} jobs are going to be launched, confirm?"):
        for barcode, image_id in tqdm.tqdm(missing_items, desc="image"):
            image_url = generate_image_url(barcode, image_id)
            enqueue_job(
                func,
                low_queue,
                job_kwargs={"result_ttl": 0},
                barcode=barcode,
                image_url=image_url,
                server_domain=settings.OFF_SERVER_DOMAIN,
                **kwargs,
            )


@app.command()
def init_elasticsearch(
    load_index: bool = False,
    load_data: bool = True,
    to_load: Optional[list[str]] = None,
) -> None:
    """
    This command is used for manual insertion of the Elasticsearch data and/or indexes
    for products and categorties.

    to_load specifies which indexes/data should be loaded - supported values are
    in robotoff.settings.ElasticsearchIndex.
    """
    from robotoff.elasticsearch.export import ElasticsearchExporter
    from robotoff.settings import ElasticsearchIndex
    from robotoff.utils import get_logger
    from robotoff.utils.es import get_es_client

    logger = get_logger()

    es_exporter = ElasticsearchExporter(get_es_client())

    if not to_load:
        return

    for item in to_load:
        if item not in ElasticsearchIndex.SUPPORTED_INDICES:
            logger.error(f"Skipping over unknown Elasticsearch type: '{item}'")
            continue
        if load_index:
            es_exporter.load_index(item, ElasticsearchIndex.SUPPORTED_INDICES[item])
        if load_data:
            es_exporter.export_index_data(item)


@app.command()
def add_logo_to_ann(sleep_time: float = 0.5) -> None:
    import time
    from itertools import groupby

    import requests
    import tqdm

    from robotoff.logos import add_logos_to_ann, get_stored_logo_ids
    from robotoff.models import ImageModel, ImagePrediction, LogoAnnotation, db
    from robotoff.utils import get_logger

    logger = get_logger()
    seen = get_stored_logo_ids()

    with db:
        logos_iter = tqdm.tqdm(
            LogoAnnotation.select()
            .join(ImagePrediction)
            .join(ImageModel)
            .where(LogoAnnotation.nearest_neighbors.is_null())
            .order_by(ImageModel.id)
            .iterator()
        )
        for _, logo_batch in groupby(logos_iter, lambda x: x.image_prediction.image.id):
            logos = list(logo_batch)

            if all(logo.id in seen for logo in logos):
                continue

            image = logos[0].image_prediction.image
            logger.info(f"Adding logos of image {image.id}")
            try:
                added = add_logos_to_ann(image, logos)
            except requests.exceptions.ReadTimeout:
                logger.warning("Request timed-out during logo addition")
                continue

            logger.info(f"Added: {added}")

            if sleep_time:
                time.sleep(sleep_time)


@app.command()
def import_logos(
    data_path: pathlib.Path,
    model_name: str = "universal-logo-detector",
    model_version: str = "tf-universal-logo-detector-1.0",
) -> None:
    from robotoff.cli.logos import insert_batch
    from robotoff.models import db
    from robotoff.utils import get_logger

    logger = get_logger()
    logger.info("Starting image prediction import...")

    with db:
        inserted = insert_batch(data_path, model_name, model_version)

    logger.info("{} image predictions inserted".format(inserted))


@app.command()
def export_logo_annotation(
    output: pathlib.Path,
    server_domain: Optional[str] = None,
    annotated: Optional[bool] = None,
) -> None:
    from robotoff.models import ImageModel, ImagePrediction, LogoAnnotation, db
    from robotoff.utils import dump_jsonl

    with db:
        where_clauses = []

        if server_domain is not None:
            where_clauses.append(ImageModel.server_domain == server_domain)

        if annotated is not None:
            where_clauses.append(LogoAnnotation.annotation_value.is_null(not annotated))

        query = LogoAnnotation.select().join(ImagePrediction).join(ImageModel)
        if where_clauses:
            query = query.where(*where_clauses)

        logo_iter = query.iterator()
        dict_iter = (logo.to_dict() for logo in logo_iter)
        dump_jsonl(output, dict_iter)


@app.command()
def export_logos_ann(
    output: pathlib.Path = typer.Argument(
        ...,
        help="Path to the output file, can either have .json or .json.gz as "
        "extension",
    ),
) -> None:
    """Export all information about logo in DB necessary to generate logo
    crops."""
    from robotoff.models import ImageModel, ImagePrediction, LogoAnnotation, db
    from robotoff.utils import dump_jsonl

    with db:
        query = (
            LogoAnnotation.select(
                LogoAnnotation.id,
                LogoAnnotation.bounding_box,
                LogoAnnotation.score,
                ImageModel.image_id,
                ImageModel.barcode,
            )
            .join(ImagePrediction)
            .join(ImageModel)
        )
        dump_jsonl(output, query.dicts().iterator())


def main() -> None:
    app()



import re
from typing import Optional, Union

from robotoff.off import normalize_tag
from robotoff.prediction.types import Prediction
from robotoff.taxonomy import get_taxonomy
from robotoff.types import PredictionType
from robotoff.utils import get_logger

from .dataclass import OCRField, OCRRegex, OCRResult, get_text

logger = get_logger(__name__)


def category_taxonomisation(lang, match) -> Optional[str]:
    """Function to match categories detected via AOP REGEX with categories
    taxonomy database. If no match is possible, we return None.
    """

    unchecked_category = lang + normalize_tag(match.group("category"))

    checked_category = get_taxonomy("category").nodes.get(unchecked_category)

    # TODO: We may want to create a utility function in Taxonomy  to match
    # also with synonyms of the category existing in the taxonomy

    if checked_category is not None:
        return checked_category.id

    return None


AOC_REGEX = {
    "fr:": [
        OCRRegex(
            # re.compile(r"(?<=appellation\s).*(?=(\scontr[ôo]l[ée]e)|(\sprot[ée]g[ée]e))"),
            re.compile(
                r"(appellation)\s*(?P<category>.+)\s*(contr[ôo]l[ée]e|prot[ée]g[ée]e)"
            ),
            field=OCRField.full_text_contiguous,
            lowercase=True,
            processing_func=category_taxonomisation,
        ),
        OCRRegex(
            re.compile(
                r"(?P<category>.+)\s*(appellation d'origine contr[ôo]l[ée]e|appellation d'origine prot[ée]g[ée]e)"
            ),
            field=OCRField.full_text_contiguous,
            lowercase=True,
            processing_func=category_taxonomisation,
        ),
    ],
    "es:": [
        OCRRegex(
            re.compile(r"(?P<category>.+)(\s*denominacion de origen protegida)"),
            field=OCRField.full_text_contiguous,
            lowercase=True,
            processing_func=category_taxonomisation,
        ),
        OCRRegex(
            re.compile(r"(denominacion de origen protegida\s*)(?P<category>.+)"),
            field=OCRField.full_text_contiguous,
            lowercase=True,
            processing_func=category_taxonomisation,
        ),
    ],
    "en:": [
        OCRRegex(
            re.compile(r"(?P<category>.+)\s*(aop|dop|pdo)"),
            field=OCRField.full_text_contiguous,
            lowercase=True,
            processing_func=category_taxonomisation,
        ),
        OCRRegex(
            re.compile(r"(aop|dop|pdo)\s*(?P<category>.+)"),
            field=OCRField.full_text_contiguous,
            lowercase=True,
            processing_func=category_taxonomisation,
        ),
    ],
}


def find_category(content: Union[OCRResult, str]) -> list[Prediction]:
    """This function returns a prediction of the product category.
    For now we are extracting categories via REGEX
    only thanks to an AOP syntax but we may find in the future
    other ways to get sure prediction of categories.
    """

    predictions = []

    for lang, regex_list in AOC_REGEX.items():
        for ocr_regex in regex_list:
            text = get_text(content, ocr_regex)

            if not text:
                continue

            for match in ocr_regex.regex.finditer(text):

                if ocr_regex.processing_func:
                    category_value = ocr_regex.processing_func(lang, match)

                if category_value is None:
                    continue

                predictions.append(
                    Prediction(
                        type=PredictionType.category,
                        value_tag=category_value,
                        predictor="regex",
                        data={"text": match.group(), "notify": ocr_regex.notify},
                        automatic_processing=False,
                    )
                )
    return predictions



import enum


class WorkerQueue(enum.Enum):
    robotoff_high = "robotoff-high"
    robotoff_low = "robotoff-low"


class ObjectDetectionModel(enum.Enum):
    nutriscore = "nutriscore"
    universal_logo_detector = "universal-logo-detector"
    nutrition_table = "nutrition-table"


@enum.unique
class PredictionType(str, enum.Enum):
    """PredictionType defines the type of the prediction.

    See `InsightType` documentation for further information about each type.
    """

    ingredient_spellcheck = "ingredient_spellcheck"
    packager_code = "packager_code"
    label = "label"
    category = "category"
    image_flag = "image_flag"
    product_weight = "product_weight"
    expiration_date = "expiration_date"
    brand = "brand"
    image_orientation = "image_orientation"
    store = "store"
    nutrient = "nutrient"
    trace = "trace"
    packaging_text = "packaging_text"
    packaging = "packaging"
    location = "location"
    nutrient_mention = "nutrient_mention"
    image_lang = "image_lang"
    nutrition_image = "nutrition_image"
    nutrition_table_structure = "nutrition_table_structure"



import datetime
import functools
import re
from typing import Optional, Union

from robotoff.prediction.types import Prediction
from robotoff.types import PredictionType

from .dataclass import OCRField, OCRRegex, OCRResult, get_text


def process_full_digits_expiration_date(match, short: bool) -> Optional[datetime.date]:
    day, month, year = match.group(1, 2, 3)

    if short:
        format_str: str = "%d/%m/%y"
    else:
        format_str = "%d/%m/%Y"

    try:
        date = datetime.datetime.strptime(
            "{}/{}/{}".format(day, month, year), format_str
        ).date()
    except ValueError:
        return None

    return date


EXPIRATION_DATE_REGEX: dict[str, OCRRegex] = {
    "full_digits_short": OCRRegex(
        re.compile(r"(?<!\d)(\d{2})[-./](\d{2})[-./](\d{2})(?!\d)"),
        field=OCRField.full_text,
        lowercase=False,
        processing_func=functools.partial(
            process_full_digits_expiration_date, short=True
        ),
    ),
    "full_digits_long": OCRRegex(
        re.compile(r"(?<!\d)(\d{2})[-./](\d{2})[-./](\d{4})(?!\d)"),
        field=OCRField.full_text,
        lowercase=False,
        processing_func=functools.partial(
            process_full_digits_expiration_date, short=False
        ),
    ),
}


def find_expiration_date(content: Union[OCRResult, str]) -> list[Prediction]:
    # Parse expiration date
    #        "À consommer de préférence avant",
    results: list[Prediction] = []

    for type_, ocr_regex in EXPIRATION_DATE_REGEX.items():
        text = get_text(content, ocr_regex)

        if not text:
            continue

        for match in ocr_regex.regex.finditer(text):
            raw = match.group(0)

            if not ocr_regex.processing_func:
                continue

            date = ocr_regex.processing_func(match)

            if date is None:
                continue

            if date.year > 2025 or date.year < 2015:
                continue

            # Format dates according to ISO 8601
            value = date.strftime("%Y-%m-%d")

            results.append(
                Prediction(
                    value=value,
                    type=PredictionType.expiration_date,
                    data={"raw": raw, "type": type_, "notify": ocr_regex.notify},
                    automatic_processing=True,
                )
            )

    return results



import datetime
import itertools
import operator
import re
from typing import Iterable, Optional

import cachetools
from flashtext import KeywordProcessor

from robotoff import settings
from robotoff.prediction.types import Prediction
from robotoff.products import ProductDataset
from robotoff.taxonomy import TaxonomyType, get_taxonomy
from robotoff.types import PredictionType
from robotoff.utils import dump_json, get_logger, load_json
from robotoff.utils.text import (
    get_lemmatizing_nlp,
    strip_accents_ascii,
    strip_consecutive_spaces,
)

logger = get_logger(__name__)

SUPPORTED_LANG = {"fr", "en", "de", "es", "it", "nl"}

WEIGHT_REGEX = re.compile(
    r"[0-9]+[,.]?[0-9]*\s?(fl oz|dl|cl|mg|ml|lbs|oz|g|kg|l)(?![a-z])"
)

LABELS_REGEX = {
    "fr": re.compile(
        r"agriculture biologique|biologique|bio|igp|aop|aoc|label rouge"
        r"|commerce équitable"
    ),
    "en": re.compile(r"organic|pgi"),
    "es": re.compile(r"ecológico|ecológica|extra|comercio justo"),
    "de": re.compile(r"bio"),
    "nl": re.compile(r"bio"),
    "it": re.compile(
        r"biologica|biologico|commercio solidale|commercio equo e solidale"
        r"|da agricoltura biologica"
    ),
}

EXTRAWORDS_REGEX = {
    "fr": re.compile(r"gourmand|délicieux"),
    "en": re.compile(r"delicious"),
}

STOP_WORDS_EXCEPTIONS = {
    "fr": {"deux", "trois", "quatre"},
    "en": {"two", "three", "four"},
    "es": {"dos", "tres", "cuatro"},
    "de": {"jahren", "jahr"},
    "it": {"anno"},
}


# During stemming/stop-word removal of category names, sometimes the processed
# string becomes too generic and adds too many false positive detections.
# For instance, in French 'premier cru' is processed into 'cru' (because
# premier is a stop word), which can be found in the names of many products
# that do not belong to the "premier cru" category
MATCH_MAPS_EXCEPTIONS = {
    TaxonomyType.category.name: {
        "fr": {
            # "pas de gluten" becomes "gluten" after lemmatization
            "gluten",
            # "charolais" is a cheese name but is much more commonly used for
            # beef
            "charolais",
            # "frais" implies "en:fresh-food", but as we currently discard
            # predictions if two matches occurs with the same product name,
            # it's better to ignore this
            "frais",
            # "premier cru" becomes "cru" after lemmatization
            "cru",
            # "dès un an" becomes "an" an after lemmatization
            "an",
        },
        "es": {
            # en:fresh-food
            "fresco",
        },
        "en": {
            # en:fresh-food
            "fresh",
        },
        "de": {
            # frish -> frischen after lemmatization (en:fresh-foods)
            "frischen",
        },
    }
}

MatchMapType = dict[str, dict[str, list[tuple[str, str]]]]


def load_resources():
    """Load and cache resources."""
    logger.info("Loading matcher resources...")
    get_processors()
    get_intersect_categories_ingredients()

    for lang in SUPPORTED_LANG:
        logger.info("Loading NLP for %s...", lang)
        get_lemmatizing_nlp(lang)


def preprocess_product_name(name: str, lang: str) -> str:
    """Preprocess product name before matching:
    - remove all weight mentions (100 g, 1l,...)
    - remove all label mentions (IGP, AOP, Label Rouge,...)
    - remove uninformative words that lower recall

    :param name: the product name
    :param lang: the language of the product name
    :return: the preprocessed product name
    """
    name = name.lower()
    name = WEIGHT_REGEX.sub("", name)  # Remove weights

    if lang in LABELS_REGEX:
        # Remove labels
        name = LABELS_REGEX[lang].sub("", name)

    if lang in EXTRAWORDS_REGEX:
        # Remove marketing words
        name = EXTRAWORDS_REGEX[lang].sub("", name)

    name = strip_consecutive_spaces(name)
    name = name.strip()
    return name


def process(text: str, lang: str) -> str:
    """Text processing pipeline for category matching.

    Both category names from the taxonomy and product names should be
    processed with this function before matching. The following steps are
    performed:
    - lowercasing
    - language-specific stop word removal
    - language-specific lookup-based lemmatization: fast and independent of
    part of speech for speed and simplicity
    - NFKD unicode normalization and accent stripping

    :param text: text to process
    :param lang: language of the text
    :return: the processed string
    """
    nlp = get_lemmatizing_nlp(lang)
    lemmas = []

    # Savigny-lès-Beaune -> Savigny lès Beaune
    text = text.lower().replace("-", " ")
    doc = nlp(text)
    lang_stop_words_exceptions = STOP_WORDS_EXCEPTIONS.get(lang, set())

    for token in doc:
        if (
            (token.is_stop and token.orth_ not in lang_stop_words_exceptions)
            or token.is_punct
            or token.is_space
        ):
            # skip token as it is uninformative
            continue
        lemmas.append(token.lemma_)

    return strip_accents_ascii(" ".join(lemmas))


def generate_match_maps(taxonomy_type: str) -> MatchMapType:
    """Return a dict mapping each supported lang to a match map dict.

    Each match map dict maps a processed string (the key) to a list of
    (value_tag, original_name) tuples.
    These keys are obtained by using all synonyms and names of `value_tag` in
    different languages and using process function (this may lead to having a
    key corresponding to more than one tag).
    Queries will then be processed through the `process` function. We have a
    match if the processed query is a key of the map.

    It provides a very fast (although not memory-efficient) way of
    performing matching.

    :param taxonomy_type: the taxonomy to build the match maps from
    :return: the match maps
    """
    taxonomy = get_taxonomy(taxonomy_type)
    match_maps_exceptions = MATCH_MAPS_EXCEPTIONS.get(taxonomy_type, {})

    store: MatchMapType = {}
    for node in taxonomy.iter_nodes():
        # node.names is a dict with one expression by language
        # node.synonyms is a dict with a list of expressions by language
        for lang in set(node.names) | set(node.synonyms):
            if lang not in SUPPORTED_LANG:
                continue
            store.setdefault(lang, {})

            names = []
            if lang in node.names:
                names.append(node.names[lang])
            names += node.synonyms.get(lang, [])
            match_exceptions = match_maps_exceptions.get(lang, set())

            # now process harvested names and keep significative ones
            for name in names:
                processed = process(name, lang)
                # Don't add it to the store if empty or if it's in the
                # exception list
                if processed and processed not in match_exceptions:
                    if processed not in store[lang]:
                        store[lang][processed] = []
                    if node.id not in (item[0] for item in store[lang][processed]):
                        store[lang].setdefault(processed, []).append((node.id, name))

    return store


def get_match_maps(taxonomy_type: str) -> MatchMapType:
    """Return match maps saved on-disk for supported language.

    See `generate_match_maps` function for more information.
    """
    return load_json(  # type: ignore
        settings.CATEGORY_MATCHER_MATCH_MAPS[taxonomy_type], compressed=True
    )


@cachetools.cached(cache=cachetools.TTLCache(maxsize=1, ttl=3600))
def get_processors() -> dict[str, KeywordProcessor]:
    """Return a dict mapping lang to flashtext KeywordProcessor used to
    perform category matching.

    This enables a fast matching of query parts against matched maps keys.
    """
    logger.info("Loading category matcher processors...")
    match_maps = get_match_maps(TaxonomyType.category.name)
    processors = {}
    for lang, items in match_maps.items():
        processor = KeywordProcessor()
        for pattern, results in items.items():
            processor.add_keyword(
                pattern,
                [(id_, pattern, category_name) for id_, category_name in results],
            )
        processors[lang] = processor
    return processors


def generate_intersect_categories_ingredients() -> dict[str, set[str]]:
    """Return all taxonomized values that are present both in the ingredient
    and in the category taxonomy.

    We use match maps from both taxonomies to find names/synonyms that are the
    same in both taxonomies.
    This function is useful to spot categories that are also ingredients.
    """
    ingredient_match_maps = get_match_maps(TaxonomyType.ingredient.name)
    category_match_maps = get_match_maps(TaxonomyType.category.name)
    matches: dict[str, set[str]] = {}
    for lang in SUPPORTED_LANG:
        matches[lang] = set()

    for lang in set(ingredient_match_maps) & set(category_match_maps) & SUPPORTED_LANG:
        for key in set(ingredient_match_maps[lang]) & set(category_match_maps[lang]):
            for node_id, _ in category_match_maps[lang][key]:
                matches[lang].add(node_id)

    return matches


@cachetools.cached(cache=cachetools.Cache(maxsize=1))
def get_intersect_categories_ingredients():
    """Return intersection between category and ingredient maps saved on-disk
    for supported language.

    See `generate_intersect_categories_ingredients` function for more
    information.
    """
    logger.info("Loading category intersection ingredient...")
    return {
        k: set(v)
        for k, v in load_json(
            settings.CATEGORY_MATCHER_INTERSECT, compressed=True
        ).items()
    }


def match(query: str, lang: str) -> list[tuple[str, str, str, str, tuple[int, int]]]:
    """Perform category matching.

    If a match is also an ingredient (see
    `generate_intersect_categories_ingredients`), we only keep the match if
    it starts at the beginning of the query. It prevents many over-matchings
    with ingredient names in the query that are not the product category.

    :param query: the unprocessed product name
    :param lang: the language of the query
    :return A list of matches (category_value_tag, matched_pattern)
    """
    processors = get_processors()
    if lang not in processors:
        return []

    processor = processors[lang]
    query = preprocess_product_name(query, lang)
    query = process(query, lang)
    flat_results = []
    for results, start_idx, end_idx in processor.extract_keywords(
        query, span_info=True
    ):
        for result in results:
            flat_results.append((result, start_idx, end_idx))

    # now filter the list to avoid some false positive
    filtered = []
    category_ingredient_intersect = get_intersect_categories_ingredients()[lang]
    for (id_, pattern, category_name), start_idx, end_idx in flat_results:
        # If the category is also an ingredient, we require the match
        # to start at the beginning of the query
        # This prevent for instance "pineapple" category to match
        # "pizza with pineapple" product name
        if id_ in category_ingredient_intersect and start_idx != 0:
            continue
        filtered.append((id_, category_name, pattern, query, (start_idx, end_idx)))
    return filtered


def predict_by_lang(product: dict) -> dict[str, list[Prediction]]:
    """Predict product categories using a matching algorithm on product names
    for all supported languages.

    :param product: the product to predict the categories from, should have at
    least `product_name_{lang}` and `languages_codes` fields
    :return: a dict mapping for each supported lang a list of detected
    category Prediction
    """
    predictions: dict[str, list[Prediction]] = {}
    for lang in set(product.get("languages_codes", [])) & SUPPORTED_LANG:
        product_name = product.get(f"product_name_{lang}")

        if not product_name:
            continue

        matches = match(product_name, lang)

        predictions[lang] = [
            Prediction(
                type=PredictionType.category,
                value_tag=value_tag,
                data={
                    "lang": lang,
                    "product_name": product_name,
                    "category_name": category_name,
                    "pattern": pattern,
                    "processed_product_name": processed_product_name,
                    "start_idx": start_idx,
                    "end_idx": end_idx,
                    "is_full_match": (
                        end_idx - start_idx == len(processed_product_name)
                    ),
                },
                automatic_processing=False,
                predictor="matcher",
            )
            for (
                value_tag,
                category_name,
                pattern,
                processed_product_name,
                (start_idx, end_idx),
            ) in matches
        ]
    return predictions


def predict(product: dict) -> list[Prediction]:
    """Predict product categories using a matching algorithm on product names.

    Flashtext always return the longuest match if several patterns match
    the query. Therefore, if more than one category is predicted, it means
    there was two non-overlapping matches: we don't return any prediction for
    that lang in that case.

    :param product: product properties
    :return: a list of category Prediction"""
    predictions_by_lang = predict_by_lang(product)
    predictions = []
    for _, lang_predictions in predictions_by_lang.items():
        predictions_by_value_tag = {
            value_tag: list(value_tag_predictions)
            for (value_tag, value_tag_predictions) in itertools.groupby(
                lang_predictions, operator.attrgetter("value_tag")
            )
        }
        if len(predictions_by_value_tag) != 1:
            # if there is more than one match (with different predicted
            # category), matches on this product name are ambiguous and we
            # discard them
            continue

        # all predictions in `lang_predictions` have the same `value_tag`,
        # take the first one
        predictions.append(lang_predictions[0])

    return predictions


def predict_from_dataset(
    dataset: ProductDataset, from_datetime: Optional[datetime.datetime] = None
) -> Iterable[Prediction]:
    """Return an iterable of category predictions, using the provided dataset.

    Args:
        dataset: a ProductDataset
        from_datetime: datetime threshold: only keep products modified after
            `from_datetime`
    """
    product_stream = (
        dataset.stream()
        .filter_nonempty_text_field("code")
        .filter_nonempty_text_field("product_name")
        .filter_empty_tag_field("categories_tags")
        .filter_nonempty_tag_field("countries_tags")
        .filter_nonempty_tag_field("languages_codes")
    )

    if from_datetime:
        product_stream = product_stream.filter_by_modified_datetime(
            from_t=from_datetime
        )

    logger.info("Performing prediction on products without categories")
    for product in product_stream.iter():
        for prediction in predict(product):
            prediction.barcode = product["code"]
            yield prediction


def dump_resource_files():
    """Generate and save on-disk match maps (ingredient and category) and the
    category-ingredient intersection."""
    for taxonomy_type in (TaxonomyType.ingredient.name, TaxonomyType.category.name):
        match_map = generate_match_maps(taxonomy_type)
        dump_json(
            settings.CATEGORY_MATCHER_MATCH_MAPS[taxonomy_type],
            match_map,
            compressed=True,
        )

    intersect = generate_intersect_categories_ingredients()
    dump_json(
        settings.CATEGORY_MATCHER_INTERSECT,
        {k: list(v) for k, v in intersect.items()},
        compressed=True,
    )



from robotoff.utils.types import JSONType

IMAGE_PREDICTION_IMPORTER_SCHEMA: JSONType = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Image Prediction Importer",
    "type": "object",
    "properties": {
        "predictions": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "server_domain": {"type": "string"},
                    "barcode": {"type": "string"},
                    "image_id": {"type": "string"},
                    "model_name": {"type": "string"},
                    "model_version": {"type": "string"},
                    "data": {"type": "object"},
                },
                "required": [
                    "barcode",
                    "image_id",
                    "model_name",
                    "model_version",
                    "data",
                ],
            },
        }
    },
    "required": ["predictions"],
}

UPDATE_LOGO_SCHEMA: JSONType = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Update Logo",
    "type": "object",
    "properties": {
        "value": {"type": ["string", "null"]},
        "type": {"type": "string", "minLength": 1},
    },
    "required": ["value", "type"],
}

PREDICT_CATEGORY_SCHEMA: JSONType = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Predict Category",
    "anyOf": [
        {
            "type": "object",
            "properties": {
                "barcode": {
                    "type": "string",
                    "minLength": 1,
                },
                "deepest_only": {
                    "type": "boolean",
                },
                "threshold": {"type": "number"},
                "predictors": {
                    "type": "array",
                    "items": {"enum": ["neural", "matcher"]},
                },
            },
            "required": ["barcode"],
        },
        {
            "type": "object",
            "properties": {
                "product": {
                    "type": "object",
                    "properties": {
                        "product_name": {
                            "type": "string",
                            "minLength": 1,
                        },
                        "ingredients_tags": {
                            "type": "array",
                            "items": {"type": "string"},
                        },
                    },
                    "required": ["product_name"],
                },
                "deepest_only": {
                    "type": "boolean",
                },
                "threshold": {"type": "number"},
                "predictors": {
                    "type": "array",
                    "items": {"enum": ["neural", "matcher"]},
                },
                "lang": {
                    "type": "string",
                    "minLength": 1,
                },
            },
            "required": ["product"],
        },
    ],
}


ANNOTATE_LOGO_SCHEMA: JSONType = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Annotate Logo",
    "type": "object",
    "properties": {
        "annotations": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "value": {
                        "type": ["string", "null"],
                    },
                    "type": {
                        "type": "string",
                        "enum": [
                            "brand",
                            "category",
                            "label",
                            "no_logo",
                            "nutritional_label",
                            "packager_code",
                            "packaging",
                            "qr_code",
                            "store",
                        ],
                    },
                    "logo_id": {"type": "integer"},
                },
                "required": ["value", "type", "logo_id"],
            },
        },
    },
    "required": ["annotations"],
}



from typing import Optional

from robotoff import settings
from robotoff.prediction.types import Prediction
from robotoff.taxonomy import Taxonomy
from robotoff.types import PredictionType
from robotoff.utils import http_session


class CategoryPrediction:
    """CategoryPrediction stores information about a category classification prediction."""

    #: threshold on the neural model confidence to automatically apply prediction
    NEURAL_CONFIDENCE_THRESHOLD = 1.1  # deactivated for now, 1.1 is above 1

    def __init__(self, category: str, confidence: float):
        self.category = category
        self.confidence = confidence

    def to_prediction(self) -> Prediction:
        """Converts this category prediction to a Prediction."""
        return Prediction(
            type=PredictionType.category,
            value_tag=self.category,
            data={"lang": "xx", "confidence": self.confidence},
            automatic_processing=self.confidence >= self.NEURAL_CONFIDENCE_THRESHOLD,
            predictor="neural",
        )

    def __eq__(self, other):
        """A CategoryPrediction is equal to another prediction when their attributes match."""
        if not isinstance(other, CategoryPrediction):
            return NotImplemented

        return self.category == other.category and self.confidence == other.confidence


class CategoryClassifier:
    """CategoryClassifier is responsible for generating predictions for a given product.

    param category_taxonomy: the Taxonomy.
        This is used to have hierarchy in order to remove parents from resulting category set.
    """

    def __init__(self, category_taxonomy: Taxonomy):
        self.taxonomy = category_taxonomy

    def predict(
        self,
        product: dict,
        deepest_only: bool = False,
        threshold: Optional[float] = None,
    ) -> list[Prediction]:
        """Returns an unordered list of category predictions for the given
        product.

        :param product: the product to predict the categories from, should
        have at least `product_name` and `ingredients_tags` fields
        :param deepest_only: controls whether the returned list should only
        contain the deepmost categories for a predicted taxonomy chain.

            For example, if we predict 'fresh vegetables' -> 'legumes' ->
            'beans' for a product,
            setting deepest_only=True will return ['beans'].
        :param threshold: the score above which we consider the category to be
        detected (default: 0.5)
        """
        if threshold is None:
            threshold = 0.5

        # model was train with product having a name
        if not product.get("product_name"):
            return []
        # ingredients are not mandatory, just insure correct type
        product.setdefault("ingredients_tags", [])

        data = {
            "signature_name": "serving_default",
            "instances": [
                {
                    "ingredient": product["ingredients_tags"],
                    "product_name": [product["product_name"]],
                }
            ],
        }

        r = http_session.post(
            f"{settings.TF_SERVING_BASE_URL}/category-classifier:predict",
            json=data,
            timeout=(3.0, 10.0),
        )
        r.raise_for_status()
        response = r.json()

        # Since we only sent one product in the query, we can be guaranteed that only one
        # prediction is returned by TF Serving.
        prediction = response["predictions"][0]

        # The response is always in the form:
        #   "predictions": [
        #     {
        #         "output_mapper_layer": [0.868871808, 0.801418602, ...],
        #         "output_mapper_layer_1": ["en:seafood", "en:fishes", ....],
        #     }
        #   ]
        #
        # where 'output_mapper_layer' is the confidence score for a prediction in descending order.
        #       'output_mapper_layer_1' is the category for the prediction score above.
        #
        # The model only returns top 50 predictions.

        category_predictions: list[CategoryPrediction] = []

        # We only consider predictions with a confidence score of `threshold` and above.
        for idx, confidence in enumerate(prediction["output_mapper_layer"]):
            if confidence >= threshold:
                category_predictions.append(
                    CategoryPrediction(
                        category=prediction["output_mapper_layer_1"][idx],
                        confidence=prediction["output_mapper_layer"][idx],
                    )
                )
            else:
                break

        if deepest_only:
            predicted_dict = {p.category: p for p in category_predictions}
            taxonomy_nodes = [self.taxonomy[p.category] for p in category_predictions]

            category_predictions = [
                predicted_dict[x.id]
                for x in self.taxonomy.find_deepest_nodes(taxonomy_nodes)
            ]

        return [
            category_prediction.to_prediction()
            for category_prediction in category_predictions
        ]



import dataclasses
import datetime
from typing import Any, Optional

from robotoff.types import PredictionType


@dataclasses.dataclass
class Prediction:
    type: PredictionType
    data: dict[str, Any] = dataclasses.field(default_factory=dict)
    value_tag: Optional[str] = None
    value: Optional[str] = None
    automatic_processing: Optional[bool] = None
    predictor: Optional[str] = None
    barcode: Optional[str] = None
    timestamp: Optional[datetime.datetime] = None
    source_image: Optional[str] = None
    server_domain: Optional[str] = None
    id: Optional[int] = None

    def to_dict(self) -> dict[str, Any]:
        return dataclasses.asdict(self, dict_factory=dict_factory)


def dict_factory(*args, **kwargs):
    d = dict(*args, **kwargs)
    for key, value in d.items():
        if isinstance(value, PredictionType):
            d[key] = value.name

    return d



import re
from typing import Optional, Union

from robotoff import settings
from robotoff.prediction.types import Prediction
from robotoff.types import PredictionType
from robotoff.utils import text_file_iter
from robotoff.utils.cache import CachedStore

from .dataclass import OCRField, OCRRegex, OCRResult, get_text
from .utils import generate_keyword_processor


def generate_trace_keyword_processor(labels: Optional[list[str]] = None):
    if labels is None:
        labels = list(text_file_iter(settings.OCR_TRACE_ALLERGEN_DATA_PATH))

    return generate_keyword_processor(labels)


TRACES_REGEX = OCRRegex(
    re.compile(
        r"(?:possibilit[ée] de traces|conditionné dans un atelier qui manipule|peut contenir(?: des traces)?|traces? [ée]ventuelles? d[e']|traces? d[e']|may contain)"
    ),
    field=OCRField.full_text_contiguous,
    lowercase=True,
)

TRACE_KEYWORD_PROCESSOR_STORE = CachedStore(
    fetch_func=generate_trace_keyword_processor, expiration_interval=None
)


def find_traces(content: Union[OCRResult, str]) -> list[Prediction]:
    predictions = []

    text = get_text(content, TRACES_REGEX)

    if not text:
        return []

    processor = TRACE_KEYWORD_PROCESSOR_STORE.get()

    for match in TRACES_REGEX.regex.finditer(text):
        prompt = match.group()
        end_idx = match.end()
        captured = text[end_idx : end_idx + 100]

        for (trace_tag, _), span_start, span_end in processor.extract_keywords(
            captured, span_info=True
        ):
            match_str = captured[span_start:span_end]
            predictions.append(
                Prediction(
                    type=PredictionType.trace,
                    value_tag=trace_tag,
                    data={"text": match_str, "prompt": prompt, "notify": False},
                )
            )

    return predictions



import re

from robotoff.settings import SPELLCHECK_PATTERNS_PATHS
from robotoff.spellcheck.base_spellchecker import BaseSpellchecker
from robotoff.spellcheck.exceptions import LanguageNotAllowedException

Patterns = dict[str, str]

VERSION = "1"


class PatternsSpellchecker(BaseSpellchecker):
    def __init__(self, lang: str = "fr"):
        self.lang = lang
        self.patterns: Patterns = {}

        try:
            patterns_path = SPELLCHECK_PATTERNS_PATHS[lang]
        except KeyError:
            raise LanguageNotAllowedException(
                f"Lang {lang} is not supported in PatternsSpellchecker."
                f" Allowed : {list(SPELLCHECK_PATTERNS_PATHS.keys())}."
            )

        with patterns_path.open() as f:
            current_pattern = None
            for line in f:
                line = line.strip().split("#")[0]
                if len(line) == 0:
                    current_pattern = None
                elif current_pattern is None:
                    current_pattern = line
                else:
                    self.patterns[line] = current_pattern

    @property
    def name(self):
        return super(PatternsSpellchecker, self).name + "__" + self.lang

    def correct(self, text: str) -> str:
        for pattern, replacement in self.patterns.items():
            text = replace_keep_case(pattern, replacement, text)
        return text

    def get_config(self):
        return {
            "version": VERSION,
            "name": self.__class__.__name__,
        }


def replace_keep_case(word, replacement, text):
    """
    Taken from https://stackoverflow.com/a/24894475/5333945
    Replace a word in a text while preserving the case.
    """

    def func(match):
        g = match.group()
        if g.islower():
            return replacement.lower()
        if g.istitle():
            return replacement.title()
        if g.isupper():
            return replacement.upper()
        return replacement

    return re.sub(word, func, text, flags=re.IGNORECASE)



import itertools
from typing import Optional

from robotoff.taxonomy import TaxonomyType, get_taxonomy

from .matcher import predict_by_lang
from .neural.category_classifier import CategoryClassifier


def predict_category(
    product: dict,
    neural_predictor: bool,
    matcher_predictor: bool,
    deepest_only: bool,
    threshold: Optional[float] = None,
) -> dict[str, list[dict]]:
    """Predict categories for a product.

    Two predictors are available:
    - the neural predictor
    - the matcher predictor

    Each predictor can be enabled using their associated flag parameter
    (`neural_predictor` and `matcher_predictor`). This function returns a dict
    where the key is the name of the predictor and the values are lists of
    predicted categories.

    :param product: the product to predict the categories from, should have at
    least `product_name` and `ingredients_tags` fields for neural predictor
    and `product_name_{lang}` and `languages_codes` for matching predictor
    :param neural_predictor: if True, add predictions of the neural predictor
    :param matcher_predictor: if True, add predictions of the matcher
    predictor
    :param deepest_only: controls whether the returned list should only
    contain the deepmost categories for a predicted taxonomy chain.
    :param threshold: the score above which we consider the category to be
    detected for neural predictor (default: 0.5)
    """
    response = {}
    taxonomy = get_taxonomy(TaxonomyType.category.name)
    if neural_predictor:
        predictions = CategoryClassifier(taxonomy).predict(
            product, deepest_only, threshold
        )
        response["neural"] = [
            {"value_tag": p.value_tag, "confidence": p.data["confidence"]}
            for p in predictions
        ]
    if matcher_predictor:
        predictions = list(
            itertools.chain.from_iterable(predict_by_lang(product).values())
        )

        if deepest_only:
            predicted_dict = {p.value_tag: p for p in predictions}
            taxonomy_nodes = [taxonomy[p.value_tag] for p in predictions]  # type: ignore

            predictions = [
                predicted_dict[x.id]
                for x in taxonomy.find_deepest_nodes(taxonomy_nodes)
            ]

        response["matcher"] = [
            {
                "value_tag": p.value_tag,
                "debug": {
                    "product_name": p.data["product_name"],
                    "lang": p.data["lang"],
                    "pattern": p.data["pattern"],
                    "processed_product_name": p.data["processed_product_name"],
                    "start_idx": p.data["start_idx"],
                    "end_idx": p.data["end_idx"],
                    "is_full_match": p.data["is_full_match"],
                    "category_name": p.data["category_name"],
                },
            }
            for p in itertools.chain.from_iterable(predict_by_lang(product).values())
        ]

    return response



import re
from typing import Iterable, Optional, Union

from flashtext import KeywordProcessor

from robotoff import settings
from robotoff.prediction.types import Prediction
from robotoff.types import PredictionType
from robotoff.utils import get_logger, text_file_iter
from robotoff.utils.cache import CachedStore

from .dataclass import OCRField, OCRRegex, OCRResult, get_text
from .utils import generate_keyword_processor

logger = get_logger(__name__)


def process_eu_bio_label_code(match) -> Optional[str]:
    country = match.group(1).lower()
    bio_code = match.group(2).replace("ö", "o").replace("ø", "o").lower()
    id_ = match.group(3).lower()

    if country == "de" and len(id_) != 3:
        return None

    return "en:{}-{}-{}".format(country, bio_code, id_)


def process_es_bio_label_code(match) -> str:
    return "en:es-eco-{}-{}".format(match.group(1), match.group(2)).lower()


EN_ORGANIC_REGEX_STR = [
    r"ingr[ée]dients?\sbiologiques?",
    r"ingr[ée]dients?\sbio[\s.,)]",
    r"agriculture ue/non ue biologique",
    r"agriculture bio(?:logique)?[\s.,)]",
    r"production bio(?:logique)?[\s.,)]",
]

LABELS_REGEX = {
    "en:organic": [
        OCRRegex(
            re.compile(r"|".join([r"(?:{})".format(x) for x in EN_ORGANIC_REGEX_STR])),
            field=OCRField.full_text_contiguous,
            lowercase=True,
        ),
    ],
    "xx-bio-xx": [
        # The negative lookbehind (?<![a-zA-Z]) is useful to avoid to match
        # strings if additional chars are before the label
        OCRRegex(
            re.compile(
                r"(?<![a-zA-Z])([A-Z]{2})[\-\s.](BIO|ÖKO|OKO|EKO|ØKO|ORG|Bio)[\-\s.](\d{2,3})"
            ),
            field=OCRField.text_annotations,
            lowercase=False,
            processing_func=process_eu_bio_label_code,
        ),
        # Spain specific regex
        OCRRegex(
            re.compile(r"(?<![a-zA-Z])ES[\-\s.]ECO[\-\s.](\d{3})[\-\s.]([A-Z]{2,3})"),
            field=OCRField.text_annotations,
            lowercase=False,
            processing_func=process_es_bio_label_code,
        ),
    ],
    "fr:ab-agriculture-biologique": [
        OCRRegex(
            re.compile(r"certifi[ée] ab[\s.,)]"),
            field=OCRField.full_text_contiguous,
            lowercase=True,
        ),
    ],
    "en:pgi": [
        OCRRegex(
            re.compile(
                r"indication g[ée]ographique prot[eé]g[eé]e|Indicazione geografica protetta|geschützte geografische angabe"
            ),
            field=OCRField.full_text_contiguous,
            lowercase=True,
        ),
        OCRRegex(
            re.compile(r"(?<!\w)(?:IGP|BGA|PGI)(?!\w)"),
            field=OCRField.full_text_contiguous,
            lowercase=False,
        ),
    ],
    "fr:label-rouge": [
        OCRRegex(
            re.compile(r"d[ée]cret du 0?5[./]01[./]07"),
            field=OCRField.full_text_contiguous,
            lowercase=True,
        ),
        OCRRegex(
            re.compile(r"(?<!\w)homologation(?: n°?)? ?la ?\d{2}\/\d{2}(?!\w)"),
            field=OCRField.full_text_contiguous,
            lowercase=True,
        ),
    ],
    "en:pdo": [
        OCRRegex(
            re.compile(r"(?<!\w)(?:PDO|AOP|DOP)(?!\w)"),
            field=OCRField.full_text_contiguous,
            lowercase=False,
        ),
        OCRRegex(
            re.compile(r"appellation d'origine prot[eé]g[eé]e"),
            field=OCRField.full_text_contiguous,
            lowercase=True,
        ),
    ],
    "fr:aoc": [
        OCRRegex(
            re.compile(r"(?<!\w)(?:AOC)(?!\w)"),
            field=OCRField.full_text_contiguous,
            lowercase=False,
        ),
    ],
    "en:nutriscore": [
        OCRRegex(
            re.compile(r"NUTRI-SCORE"),
            field=OCRField.full_text,
            lowercase=False,
        ),
    ],
    "en:eu-non-eu-agriculture": [
        OCRRegex(
            re.compile(
                r"agriculture ue\s?/\s?non\s?(?:-\s?)?ue|eu\s?/\s?non\s?(?:-\s?)?eu agriculture"
            ),
            field=OCRField.full_text_contiguous,
            lowercase=True,
        ),
    ],
    "en:eu-agriculture": [
        # The negative lookafter/lookbehind forbid matching "agriculture ue/non ue"
        OCRRegex(
            re.compile(r"agriculture ue(?!\s?/)|(?<!-)\s?eu agriculture"),
            field=OCRField.full_text_contiguous,
            lowercase=True,
        ),
    ],
    "en:non-eu-agriculture": [
        OCRRegex(
            re.compile(r"agriculture non\s?(?:-\s?)?ue|non\s?(?:-\s?)?eu agriculture"),
            field=OCRField.full_text_contiguous,
            lowercase=True,
        ),
    ],
    "en:no-preservatives": [
        OCRRegex(
            re.compile(
                r"senza conservanti(?! arti)|без консервантов|conserveermiddelvrij|(?<!\w)(?:sans|ni) conservateur(?!s? arti)|fără conservanți|no preservative|sin conservante(?!s? arti)|ohne konservierungsstoffe"
            ),
            field=OCRField.full_text_contiguous,
            lowercase=True,
        ),
    ],
    "en:no-flavors": [
        OCRRegex(
            re.compile(
                r"без ароматизаторов|senza aromi|zonder toegevoegde smaakstoffen|(?<!\w)(?:sans|ni) ar[ôo]mes? ajout[ée]s|sin aromas?|ohne zusatz von aromen|no flavors?"
            ),
            field=OCRField.full_text_contiguous,
            lowercase=True,
        ),
    ],
    "en:no-artificial-flavors": [
        OCRRegex(
            re.compile(
                r"без искусственных ароматизаторов|ohne künstliche aromen|sin aromas? artificiales?|vrij van kunstmatige smaakstoffen|(?<!\w)(?:sans|ni) ar[ôo]mes? artificiels?|no artificial flavors?"
            ),
            field=OCRField.full_text_contiguous,
            lowercase=True,
        ),
    ],
    "en:no-colorings": [
        OCRRegex(
            re.compile(
                r"no colorings?|no colourants?|ohne farbstoffzusatz|(?<!\w)(?:sans|ni) colorants?|zonder kleurstoffen|sin colorantes?|без красителей|senza coloranti"
            ),
            field=OCRField.full_text_contiguous,
            lowercase=True,
        ),
    ],
    "en:no-additives": [
        OCRRegex(
            re.compile(
                r"zonder toevoegingen|sin aditivos(?! arti)|(?<!\w)(?:sans|ni) additif(?!s? arti)|ohne zusätze|no additives?"
            ),
            field=OCRField.full_text_contiguous,
            lowercase=True,
        ),
    ],
}


def get_logo_annotation_labels() -> dict[str, str]:
    labels: dict[str, str] = {}

    for item in text_file_iter(settings.OCR_LOGO_ANNOTATION_LABELS_DATA_PATH):
        if "||" in item:
            logo_description, label_tag = item.split("||")
        else:
            logger.warning("'||' separator expected!")
            continue

        labels[logo_description] = label_tag

    return labels


def generate_label_keyword_processor(labels: Optional[Iterable[str]] = None):
    if labels is None:
        labels = text_file_iter(settings.OCR_LABEL_FLASHTEXT_DATA_PATH)

    return generate_keyword_processor(labels)


def extract_label_flashtext(processor: KeywordProcessor, text: str) -> list[Prediction]:
    predictions = []

    for (label_tag, _), span_start, span_end in processor.extract_keywords(
        text, span_info=True
    ):
        match_str = text[span_start:span_end]
        predictions.append(
            Prediction(
                type=PredictionType.label,
                value_tag=label_tag,
                automatic_processing=False,
                predictor="flashtext",
                data={"text": match_str, "notify": False},
            )
        )

    return predictions


LOGO_ANNOTATION_LABELS: dict[str, str] = get_logo_annotation_labels()
LABEL_KEYWORD_PROCESSOR_STORE = CachedStore(
    fetch_func=generate_label_keyword_processor, expiration_interval=None
)


def find_labels(content: Union[OCRResult, str]) -> list[Prediction]:
    predictions = []

    for label_tag, regex_list in LABELS_REGEX.items():
        for ocr_regex in regex_list:
            text = get_text(content, ocr_regex)

            if not text:
                continue

            for match in ocr_regex.regex.finditer(text):
                if ocr_regex.processing_func:
                    label_value = ocr_regex.processing_func(match)

                    if label_value is None:
                        continue

                else:
                    label_value = label_tag

                predictions.append(
                    Prediction(
                        type=PredictionType.label,
                        value_tag=label_value,
                        predictor="regex",
                        data={"text": match.group(), "notify": ocr_regex.notify},
                    )
                )

    processor = LABEL_KEYWORD_PROCESSOR_STORE.get()

    text = get_text(content)
    predictions += extract_label_flashtext(processor, text)

    if isinstance(content, OCRResult):
        for logo_annotation in content.logo_annotations:
            if logo_annotation.description in LOGO_ANNOTATION_LABELS:
                label_tag = LOGO_ANNOTATION_LABELS[logo_annotation.description]

                predictions.append(
                    Prediction(
                        type=PredictionType.label,
                        value_tag=label_tag,
                        automatic_processing=False,
                        predictor="google-cloud-vision",
                        data={"confidence": logo_annotation.score},
                    )
                )

    return predictions



import json
import pathlib
from typing import Callable, Iterable, Optional, TextIO, Union

import requests

from robotoff.off import generate_json_ocr_url, get_barcode_from_path, split_barcode
from robotoff.prediction.types import Prediction
from robotoff.settings import BaseURLProvider
from robotoff.types import PredictionType
from robotoff.utils import get_logger, http_session, jsonl_iter, jsonl_iter_fp
from robotoff.utils.types import JSONType

from .brand import find_brands
from .category import find_category
from .dataclass import OCRParsingException, OCRResult, OCRResultGenerationException
from .expiration_date import find_expiration_date
from .image_flag import flag_image
from .image_lang import get_image_lang
from .image_orientation import find_image_orientation
from .label import find_labels
from .location import find_locations
from .nutrient import find_nutrient_mentions, find_nutrient_values
from .packager_code import find_packager_codes
from .product_weight import find_product_weight
from .store import find_stores
from .trace import find_traces

logger = get_logger(__name__)


PREDICTION_TYPE_TO_FUNC: dict[
    str, Callable[[Union[OCRResult, str]], list[Prediction]]
] = {
    PredictionType.category: find_category,
    PredictionType.packager_code: find_packager_codes,
    PredictionType.label: find_labels,
    PredictionType.expiration_date: find_expiration_date,
    PredictionType.image_flag: flag_image,
    PredictionType.image_orientation: find_image_orientation,
    PredictionType.product_weight: find_product_weight,
    PredictionType.trace: find_traces,
    PredictionType.nutrient: find_nutrient_values,
    PredictionType.nutrient_mention: find_nutrient_mentions,
    PredictionType.brand: find_brands,
    PredictionType.store: find_stores,
    PredictionType.location: find_locations,
    PredictionType.image_lang: get_image_lang,
}


def fetch_images_for_ean(ean: str):
    url = BaseURLProvider().get() + "/api/v0/product/{}.json?fields=images".format(ean)
    images = http_session.get(url).json()
    return images


def get_json_for_image(barcode: str, image_id: str) -> Optional[JSONType]:
    url = generate_json_ocr_url(barcode, image_id)
    r = http_session.get(url)

    if r.status_code == 404:
        return None

    return r.json()


def get_ocr_result(
    ocr_url: str, session: requests.Session, error_raise: bool = True
) -> Optional[OCRResult]:
    try:
        r = session.get(ocr_url)
    except requests.exceptions.RequestException as e:
        error_message = "HTTP Error when fetching OCR URL"
        if error_raise:
            raise OCRResultGenerationException(error_message, ocr_url) from e

        logger.warning(error_message + ": %s", ocr_url, exc_info=e)
        return None

    try:
        ocr_data: dict = r.json()
    except json.JSONDecodeError as e:
        error_message = "Error while decoding OCR JSON"
        if error_raise:
            raise OCRResultGenerationException(error_message, ocr_url) from e

        logger.warning(error_message + ": %s", ocr_url, exc_info=e)
        return None

    try:
        return OCRResult.from_json(ocr_data)
    except OCRParsingException as e:
        if error_raise:
            raise OCRResultGenerationException(str(e), ocr_url) from e

        logger.warning("Error while parsing OCR JSON from %s", ocr_url, exc_info=e)
        return None


def extract_predictions(
    content: Union[OCRResult, str],
    prediction_type: PredictionType,
    barcode: Optional[str] = None,
    source_image: Optional[str] = None,
) -> list[Prediction]:
    """Extract predictions from OCR using for provided prediction type.

    :param content: OCR output to extract predictions from.
    :param barcode: Barcode to add to each prediction, defaults to None.
    :param source_image: `source_image`to add to each prediction, defaults to
    None.
    :return: The generated predictions.
    """
    if prediction_type in PREDICTION_TYPE_TO_FUNC:
        predictions = PREDICTION_TYPE_TO_FUNC[prediction_type](content)
        for prediction in predictions:
            prediction.barcode = barcode
            prediction.source_image = source_image
        return predictions
    else:
        raise ValueError(f"unknown prediction type: {prediction_type}")


def is_barcode(text: str):
    return text.isdigit()


def get_source(
    image_name: str, json_path: Optional[str] = None, barcode: Optional[str] = None
) -> str:
    if not barcode:
        barcode = get_barcode_from_path(str(json_path))

        if not barcode:
            raise ValueError("invalid JSON path: {}".format(json_path))

    return "/{}/{}.jpg" "".format("/".join(split_barcode(barcode)), image_name)


def ocr_content_iter(items: Iterable[JSONType]) -> Iterable[tuple[Optional[str], dict]]:
    for item in items:
        if "content" in item:
            source = item["source"].replace("//", "/").replace(".json", ".jpg")
            yield source, item["content"]


def ocr_iter(
    source: Union[str, TextIO, pathlib.Path]
) -> Iterable[tuple[Optional[str], dict]]:
    if isinstance(source, pathlib.Path):
        items = jsonl_iter(source)
        yield from ocr_content_iter(items)

    elif not isinstance(source, str):
        items = jsonl_iter_fp(source)
        yield from ocr_content_iter(items)



import gettext
from typing import Optional

from robotoff import settings


class TranslationStore:
    def __init__(self):
        self.translations: dict[str, gettext.NullTranslations] = {}

    def load(self):
        for lang_dir in settings.I18N_DIR.glob("*"):
            if not lang_dir.is_dir():
                continue

            lang = lang_dir.name
            t = gettext.translation(
                "robotoff", str(settings.I18N_DIR), languages=[lang]
            )
            self.translations[lang] = t

    def get(self, lang: str) -> Optional[gettext.NullTranslations]:
        return self.translations.get(lang)

    def gettext(self, lang: str, message: str) -> str:
        translation = self.translations.get(lang)

        if translation is None:
            return message

        return translation.gettext(message)



import json
from typing import Iterable

import elasticsearch

from robotoff import settings


def get_es_client():
    return elasticsearch.Elasticsearch(settings.ELASTICSEARCH_HOSTS)


def perform_export(
    client, data: Iterable[tuple[str, dict]], index: str, batch_size=100
) -> int:
    batch = []
    rows_inserted = 0

    for id_, item in data:
        batch.append(({"index": {"_id": id_}}, item))

        if len(batch) >= batch_size:
            insert_batch(client, batch, index)
            rows_inserted += len(batch)
            batch = []

    if batch:
        insert_batch(client, batch, index)
        rows_inserted += len(batch)

    return rows_inserted


def insert_batch(client, batch: Iterable[tuple[dict, dict]], index: str):
    body = ""
    for action, source in batch:
        body += "{}\n{}\n".format(json.dumps(action), json.dumps(source))

    client.bulk(body=body, index=index, doc_type=settings.ELASTICSEARCH_TYPE)


def generate_msearch_body(index: str, queries: Iterable[dict]):
    lines = []

    for query in queries:
        lines.append(json.dumps({"index": index}))
        lines.append(json.dumps(query))

    return "\n".join(lines)



from robotoff.spellcheck.exceptions import TokenLengthMismatchException
from robotoff.spellcheck.items import AtomicCorrection, Offset


class CorrectionFormatter:
    def format(
        self,
        original_tokens: list[dict],
        suggestion_tokens: list[dict],
        offset: Offset,
        score: int,
    ) -> list[AtomicCorrection]:
        if len(original_tokens) != len(suggestion_tokens):
            raise TokenLengthMismatchException()

        atomic_corrections = []
        for original_token, suggestion_token in zip(original_tokens, suggestion_tokens):
            original_token_str = original_token["token"]
            suggestion_token_str = suggestion_token["token"]

            if original_token_str.lower() != suggestion_token_str:
                if original_token_str.isupper():
                    token_str = suggestion_token_str.upper()
                elif original_token_str.istitle():
                    token_str = suggestion_token_str.capitalize()
                else:
                    token_str = suggestion_token_str

                atomic_corrections.append(
                    AtomicCorrection(
                        original=original_token_str,
                        correction=token_str,
                        offset=Offset(
                            offset.start + original_token["start_offset"],
                            offset.start + original_token["end_offset"],
                        ),
                        score=score,
                    )
                )
        return atomic_corrections



import re
from typing import Optional

from robotoff.spellcheck.base_spellchecker import BaseSpellchecker
from robotoff.spellcheck.vocabulary.utils import Vocabulary
from robotoff.utils.text import get_blank_nlp

TOKENS = list[str]
ADDITIVES_REGEX = re.compile(r"(?:E ?\d{3,5}[a-z]*)", re.IGNORECASE)
VERSION = "1"


class VocabularySpellchecker(BaseSpellchecker):
    def __init__(self):
        self.wikipedia_voc = Vocabulary("wikipedia_lower")
        self.ingredients_voc = Vocabulary("ingredients_fr_tokens") | Vocabulary(
            "ingredients_fr"
        )

    def correct(self, text: str) -> str:
        for token in self.tokenize(text, remove_additives=True):
            if all(c.isalpha() for c in token):
                if token not in self.wikipedia_voc:
                    suggestion: Optional[str] = self.ingredients_voc.suggest(token)
                    if suggestion is not None:
                        text = text.replace(token, suggestion)
        return text

    @staticmethod
    def tokenize(text: str, remove_additives: bool = False) -> TOKENS:
        tokens: TOKENS = []

        nlp = get_blank_nlp("fr")
        for token in nlp(text):
            tokens.append(token.orth_)
        tokens = [token for token in tokens if any(c.isalnum() for c in token)]
        if remove_additives:
            tokens = [token for token in tokens if ADDITIVES_REGEX.match(token) is None]
        return tokens

    def get_config(self):
        return {
            "version": VERSION,
            "name": self.__class__.__name__,
        }



from robotoff.spellcheck.base_spellchecker import BaseSpellchecker
from robotoff.spellcheck.elasticsearch.correction_formatter import CorrectionFormatter
from robotoff.spellcheck.elasticsearch.es_handler import ElasticsearchHandler
from robotoff.spellcheck.exceptions import TokenLengthMismatchException
from robotoff.spellcheck.items import (
    AtomicCorrection,
    Ingredients,
    SpellcheckItem,
    SpellcheckIteration,
)

VERSION = "1"


class ElasticSearchSpellchecker(BaseSpellchecker):
    def __init__(self, client, **kwargs):
        self.client = client
        self.kwargs = kwargs

    @property
    def name(self) -> str:
        return super(ElasticSearchSpellchecker, self).name + "".join(
            [f"__{kw}_{value}" for kw, value in self.kwargs.items()]
        )

    def predict_one(self, item: SpellcheckItem) -> SpellcheckItem:
        if item.is_lang_allowed:
            original = item.latest_correction
            atomic_corrections = self._process(original)
            item.iterations.append(
                SpellcheckIteration(
                    model=self.name,
                    original=original,
                    atomic_corrections=atomic_corrections,
                )
            )
        return item

    def correct(self, text: str) -> str:
        return SpellcheckIteration(
            original=text, atomic_corrections=self._process(text)
        ).corrected_text

    def _process(self, text: str) -> list[AtomicCorrection]:
        es_handler = ElasticsearchHandler(self.client, **self.kwargs)
        correction_formatter = CorrectionFormatter()

        ingredients: Ingredients = Ingredients.from_text(text)
        suggestions = es_handler.suggest_batch(ingredients.get_iter())

        corrections = []
        for idx, suggestion in enumerate(suggestions):
            if not suggestion.get("options"):
                continue
            main_option = suggestion["options"][0]

            try:
                corrections.extend(
                    correction_formatter.format(
                        original_tokens=es_handler.analyze(
                            ingredients.get_normalized_ingredient_text(idx)
                        ),
                        suggestion_tokens=es_handler.analyze(main_option["text"]),
                        offset=ingredients.offsets[idx],
                        score=main_option["score"],
                    )
                )
            except TokenLengthMismatchException:
                continue

        for correction in corrections:
            correction.model = self.name

        return corrections

    def get_config(self):
        return {
            "version": VERSION,
            "name": self.__class__.__name__,
            **self.kwargs,
        }



import json
import operator
from typing import Optional

import requests

from robotoff import settings
from robotoff.insights.dataclass import InsightType
from robotoff.logo_label_type import LogoLabelType
from robotoff.models import LogoAnnotation, ProductInsight, crop_image_url
from robotoff.prediction.types import Prediction
from robotoff.utils import get_logger, http_session
from robotoff.utils.types import JSONType

logger = get_logger(__name__)


class SlackException(Exception):
    pass


class NotifierInterface:
    """NotifierInterface is an interface for posting
    Robotoff-related alerts and notifications
    to various channels.
    """

    # Note: we do not use abstract methods,
    # for a notifier might choose to only implements a few

    def notify_image_flag(
        self, predictions: list[Prediction], source_image: str, barcode: str
    ):
        pass

    def notify_automatic_processing(self, insight: ProductInsight):
        pass

    def send_logo_notification(
        self, logo: LogoAnnotation, probs: dict[LogoLabelType, float]
    ):
        pass


class NotifierFactory:
    """NotifierFactory is responsible for creating a notifier to post notifications to."""

    @staticmethod
    def get_notifier() -> NotifierInterface:
        notifiers: list[NotifierInterface] = []
        token = settings.slack_token()
        if token == "":
            # use a Noop notifier to get logs for tests and dev
            notifiers.append(NoopSlackNotifier())
        else:
            notifiers.append(SlackNotifier(token))
        moderation_service_url: Optional[str] = settings.IMAGE_MODERATION_SERVICE_URL
        if moderation_service_url:
            notifiers.append(ImageModerationNotifier(moderation_service_url))
        if len(notifiers) == 1:
            return notifiers[0]
        else:
            return MultiNotifier(notifiers)


def _sensitive_image(flag_type: str, flagged_label: str) -> bool:
    """Determines whether the given flagged image should be considered as sensitive."""
    is_human: bool = flagged_label in {
        "face",
        "head",
        "selfie",
        "hair",
        "forehead",
        "chin",
        "cheek",
        "tooth",
        "eyebrow",
        "ear",
        "neck",
        "jaw",
        "nose",
        "facial expression",
        "glasses",
        "eyewear",
        "child",
        "baby",
        "human",
    }
    return (
        is_human and flag_type == "label_annotation"
    ) or flag_type == "safe_search_annotation"


def _slack_message_block(
    message_text: str, with_image: Optional[str] = None
) -> list[dict]:
    """Formats given parameters into a Slack message block."""
    block = {
        "type": "section",
        "text": {
            "type": "mrkdwn",
            "text": message_text,
        },
    }

    if with_image:
        block["accessory"] = {
            "type": "image",
            "image_url": with_image,
            "alt_text": "-",
        }
    return [block]


class MultiNotifier(NotifierInterface):
    """Aggregate multiple notifiers in one instance

    See NotifierInterface for methods documentation

    :param notifiers: the notifiers to dispatch to
    """

    def __init__(self, notifiers: list[NotifierInterface]):
        self.notifiers: list[NotifierInterface] = notifiers

    def _dispatch(self, function_name: str, *args, **kwargs):
        """dispatch call to function_name to all notifiers"""
        for notifier in self.notifiers:
            fn = getattr(notifier, function_name)
            fn(*args, **kwargs)

    def notify_image_flag(
        self, predictions: list[Prediction], source_image: str, barcode: str
    ):
        self._dispatch("notify_image_flag", predictions, source_image, barcode)

    def notify_automatic_processing(self, insight: ProductInsight):
        self._dispatch("notify_automatic_processing", insight)

    def send_logo_notification(
        self, logo: LogoAnnotation, probs: dict[LogoLabelType, float]
    ):
        self._dispatch("send_logo_notification", logo, probs)


class ImageModerationNotifier(NotifierInterface):
    """Notifier to dispatch to image moderation server

    :param service_url: base url for image moderation service
    """

    def __init__(self, service_url):
        self.service_url = service_url.rstrip("/")

    def notify_image_flag(
        self, predictions: list[Prediction], source_image: str, barcode: str
    ):
        """Send image to the moderation server so that a human can moderate it"""
        if not predictions:
            return
        image_url = f"{settings.OFF_IMAGE_BASE_URL}/{source_image.lstrip('/')}"
        image_id = int(source_image.rsplit("/", 1)[-1].split(".", 1)[0])
        params = {"imgid": image_id, "url": image_url}
        try:
            http_session.put(f"{self.service_url}/{barcode}", data=params)
        except Exception:
            logger.exception(
                "Error while notifying image to moderation service",
                extra={"params": params, "url": image_url, "barcode": barcode},
            )


class SlackNotifier(NotifierInterface):
    """Notifier to send messages on specific slack channels"""

    # Slack channel IDs.
    ROBOTOFF_ALERT_CHANNEL = "CGKPALRCG"  # robotoff-alerts-annotations
    ROBOTOFF_PRIVATE_IMAGE_ALERT_CHANNEL = "GGMRWLEF2"  # moderation-off-alerts-private
    ROBOTOFF_PUBLIC_IMAGE_ALERT_CHANNEL = "CT2N423PA"  # moderation-off-alerts

    BASE_URL = "https://slack.com/api"
    POST_MESSAGE_URL = BASE_URL + "/chat.postMessage"
    COLLAPSE_LINKS_PARAMS = {
        "unfurl_links": False,
        "unfurl_media": False,
    }

    def __init__(self, slack_token: str):
        """Should not be called directly, use the NotifierFactory instead."""
        self.slack_token = slack_token

    def notify_image_flag(
        self, predictions: list[Prediction], source_image: str, barcode: str
    ):
        """Sends alerts to Slack channels for flagged images."""
        if not predictions:
            return

        text = ""
        slack_channel: str = self.ROBOTOFF_PUBLIC_IMAGE_ALERT_CHANNEL

        for flagged in predictions:
            flag_type = flagged.data["type"]
            label = flagged.data["label"]

            if _sensitive_image(flag_type, label):
                slack_channel = self.ROBOTOFF_PRIVATE_IMAGE_ALERT_CHANNEL

            if flag_type in ("safe_search_annotation", "label_annotation"):
                likelihood = flagged.data["likelihood"]
                text += f"type: {flag_type}\nlabel: *{label}*, score: {likelihood}\n"
            else:
                match_text = flagged.data["text"]
                text += f"type: {flag_type}\nlabel: *{label}*, match: {match_text}\n"

        edit_url = f"{settings.BaseURLProvider().get()}/cgi/product.pl?type=edit&code={barcode}"
        image_url = settings.OFF_IMAGE_BASE_URL + source_image

        full_text = f"{text}\n <{image_url}|Image> -- <{edit_url}|*Edit*>"
        message = _slack_message_block(full_text, with_image=image_url)

        self._post_message(message, slack_channel, **self.COLLAPSE_LINKS_PARAMS)

    def notify_automatic_processing(self, insight: ProductInsight):
        product_url = f"{settings.BaseURLProvider().get()}/product/{insight.barcode}"

        if insight.source_image:
            if insight.data and "bounding_box" in insight.data:
                image_url = crop_image_url(
                    insight.source_image, insight.data.get("bounding_box")
                )
            else:
                image_url = f"{settings.BaseURLProvider().static().get()}/images/products{insight.source_image}"
            metadata_text = f"(<{product_url}|product>, <{image_url}|source image>)"
        else:
            metadata_text = f"(<{product_url}|product>)"

        value = insight.value or insight.value_tag

        if insight.type in {
            InsightType.product_weight.name,
            InsightType.expiration_date.name,
        }:
            text = f"The {insight.type} `{value}` (match: `{insight.data['raw']}`) was automatically added to product {insight.barcode}"
        else:
            text = f"The `{value}` {insight.type} was automatically added to product {insight.barcode}"

        message = _slack_message_block(f"{text} {metadata_text}")
        self._post_message(
            message, self.ROBOTOFF_ALERT_CHANNEL, **self.COLLAPSE_LINKS_PARAMS
        )

    def _get_base_params(self) -> JSONType:
        return {
            "username": "robotoff-bot",
            "token": self.slack_token,
            "icon_url": "https://s3-us-west-2.amazonaws.com/slack-files2/"
            "bot_icons/2019-03-01/565595869687_48.png",
        }

    def send_logo_notification(
        self, logo: LogoAnnotation, probs: dict[LogoLabelType, float]
    ):
        crop_url = logo.get_crop_image_url()
        prob_text = "\n".join(
            (
                f"{label[0]} - {label[1]}: {prob:.2g}"
                for label, prob in sorted(
                    probs.items(), key=operator.itemgetter(1), reverse=True
                )
            )
        )
        barcode = logo.image_prediction.image.barcode
        base_off_url = settings.BaseURLProvider().get()
        text = (
            f"Prediction for <{crop_url}|image> "
            f"(<https://hunger.openfoodfacts.org/logos?logo_id={logo.id}|annotate logo>, "
            f"<{base_off_url}/product/{barcode}|product>):\n{prob_text}"
        )
        self._post_message(_slack_message_block(text), self.ROBOTOFF_ALERT_CHANNEL)

    def _post_message(
        self,
        blocks: list[dict],
        channel: str,
        **kwargs,
    ):
        try:
            params: JSONType = {
                **(self._get_base_params()),
                "channel": channel,
                "blocks": json.dumps(blocks),
                **kwargs,
            }

            r = http_session.post(self.POST_MESSAGE_URL, data=params)
            response_json = _get_slack_json(r)
            return response_json
        except Exception as e:
            logger.error(
                "An exception occurred when sending a Slack notification", exc_info=e
            )


class NoopSlackNotifier(SlackNotifier):
    """NoopSlackNotifier is a NOOP SlackNotifier used in dev/local executions of Robotoff."""

    def __init__(self):
        super().__init__("")

    def _post_message(
        self,
        blocks: list[dict],
        channel: str,
        **kwargs,
    ):
        """Overrides the actual posting to Slack with logging of the args that would've been posted."""
        logger.info(
            f"Alerting on slack channel '{channel}', with message:\n{blocks}\nand additional args:\n{kwargs}"
        )


def _get_slack_json(response: requests.Response) -> JSONType:
    json_data = response.json()

    if not response.ok:
        raise SlackException(
            "Non-200 status code from Slack: "
            "{}, response: {}"
            "".format(response.status_code, json_data)
        )

    if not json_data.get("ok", False):
        raise SlackException("Non-ok response: {}".format(json_data))

    return json_data



from typing import Iterable

from robotoff import settings
from robotoff.utils.es import generate_msearch_body


class ElasticsearchHandler:
    def __init__(
        self,
        client,
        confidence: float = 1.0,
        size: int = 1,
        min_word_length: int = 4,
        suggest_mode: str = "missing",
        suggester_name: str = "autocorrect",
        reverse: bool = True,
        index_name: str = settings.ElasticsearchIndex.PRODUCT,
    ):
        self.client = client
        self.confidence = confidence
        self.size = size
        self.min_word_length = min_word_length
        self.suggest_mode = suggest_mode
        self.suggester_name = suggester_name
        self.reverse = reverse
        self.index_name = index_name

    def analyze(self, text: str):
        return self.client.indices.analyze(
            index=settings.ElasticsearchIndex.PRODUCT,
            body={"tokenizer": "standard", "text": text},
        )["tokens"]

    def suggest(self, text: str) -> dict:
        return self.suggest_batch([text])[0]

    def suggest_batch(self, texts: Iterable[str]) -> list[dict]:
        queries = [self.__generate_query(text) for text in texts]
        body = generate_msearch_body(self.index_name, queries)
        response = self.client.msearch(body=body)
        suggestions = self.__postprocess_response(response)
        return suggestions

    def __generate_query(self, text):
        direct_generators = [
            {
                "field": "ingredients_text_fr.trigram",
                "suggest_mode": self.suggest_mode,
                "min_word_length": self.min_word_length,
            }
        ]

        if self.reverse:
            direct_generators.append(
                {
                    "field": "ingredients_text_fr.reverse",
                    "suggest_mode": self.suggest_mode,
                    "min_word_length": self.min_word_length,
                    "pre_filter": "reverse",
                    "post_filter": "reverse",
                },
            )
        return {
            "suggest": {
                "text": text,
                self.suggester_name: {
                    "phrase": {
                        "confidence": self.confidence,
                        "field": "ingredients_text_fr.trigram",
                        "size": self.size,
                        "gram_size": 3,
                        "direct_generator": direct_generators,
                        "smoothing": {"laplace": {"alpha": 0.5}},
                    }
                },
            }
        }

    def __postprocess_response(self, response):
        suggestions = []
        for r in response["responses"]:
            if r["status"] != 200:
                root_cause = r["error"]["root_cause"][0]
                error_type = root_cause["type"]
                error_reason = root_cause["reason"]
                print(f"Elasticsearch error: {error_reason} [{error_type}]")
                continue
            suggestions.append(r["suggest"][self.suggester_name][0])
        return suggestions



import csv
import datetime
import functools
import hashlib
import io
import tempfile
import uuid
from typing import Optional

import falcon
import orjson
import peewee
import requests
from falcon.media.validators import jsonschema
from falcon_cors import CORS
from falcon_multipart.middleware import MultipartMiddleware
from PIL import Image
from sentry_sdk.integrations.falcon import FalconIntegration

from robotoff import settings
from robotoff.app import schema
from robotoff.app.auth import BasicAuthDecodeError, basic_decode
from robotoff.app.core import (
    SkipVotedOn,
    SkipVotedType,
    get_image_predictions,
    get_images,
    get_insights,
    get_logo_annotation,
    get_predictions,
    save_annotation,
)
from robotoff.app.middleware import DBConnectionMiddleware
from robotoff.insights.extraction import (
    DEFAULT_OCR_PREDICTION_TYPES,
    extract_ocr_predictions,
)
from robotoff.insights.question import QuestionFormatter, QuestionFormatterFactory
from robotoff.logos import generate_insights_from_annotated_logos
from robotoff.models import (
    ImageModel,
    ImagePrediction,
    LogoAnnotation,
    ProductInsight,
    batch_insert,
)
from robotoff.off import (
    OFFAuthentication,
    generate_image_path,
    get_barcode_from_url,
    get_product,
    get_server_type,
)
from robotoff.prediction.category import predict_category
from robotoff.prediction.object_detection import ObjectDetectionModelRegistry
from robotoff.prediction.ocr.dataclass import OCRParsingException
from robotoff.products import get_product_dataset_etag
from robotoff.spellcheck import SPELLCHECKERS, Spellchecker
from robotoff.taxonomy import is_prefixed_value, match_taxonomized_value
from robotoff.types import PredictionType
from robotoff.utils import get_image_from_url, get_logger, http_session
from robotoff.utils.es import get_es_client
from robotoff.utils.i18n import TranslationStore
from robotoff.utils.text import get_tag
from robotoff.utils.types import JSONType
from robotoff.workers.queues import enqueue_in_job, enqueue_job, high_queue, low_queue
from robotoff.workers.tasks import (
    delete_product_insights_job,
    download_product_dataset_job,
    run_import_image_job,
    update_insights_job,
)

logger = get_logger()

settings.init_sentry(integrations=[FalconIntegration()])

es_client = get_es_client()

TRANSLATION_STORE = TranslationStore()
TRANSLATION_STORE.load()


def _get_skip_voted_on(
    auth: Optional[OFFAuthentication], device_id: str
) -> SkipVotedOn:
    """Helper function for constructing SkipVotedOn objects based on request params."""
    if not auth:
        return SkipVotedOn(SkipVotedType.DEVICE_ID, device_id)

    username: Optional[str] = auth.get_username()
    if not username:
        return SkipVotedOn(SkipVotedType.DEVICE_ID, device_id)

    return SkipVotedOn(SkipVotedType.USERNAME, username)


###########
# IMPORTANT: remember to update documentation at doc/references/api.yml if you change API
###########


class ProductInsightResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response, barcode: str):
        server_domain: Optional[str] = req.get_param("server_domain")
        response: JSONType = {}
        insights = [
            insight.to_dict()
            for insight in get_insights(
                barcode=barcode, server_domain=server_domain, limit=None
            )
        ]

        if not insights:
            response["status"] = "no_insights"
        else:
            response["insights"] = insights
            response["status"] = "found"

        resp.media = response


class ProductInsightDetail:
    def on_get(self, req: falcon.Request, resp: falcon.Response, insight_id: uuid.UUID):
        try:
            insight: ProductInsight = ProductInsight.get_by_id(insight_id)
        except ProductInsight.DoesNotExist:
            raise falcon.HTTPNotFound()

        resp.media = insight.to_dict()


class InsightCollection:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        response: JSONType = {}
        count: int = req.get_param_as_int("count", min_value=1, default=25)
        page: int = req.get_param_as_int("page", min_value=1, default=1)
        keep_types: Optional[list[str]] = req.get_param_as_list(
            "insight_types", required=False
        )
        barcode: Optional[str] = req.get_param("barcode")
        country: Optional[str] = req.get_param("country")
        annotated: Optional[bool] = req.get_param_as_bool("annotated")
        annotation: Optional[int] = req.get_param_as_int("annotation")
        value_tag: str = req.get_param("value_tag")
        brands = req.get_param_as_list("brands") or None
        predictor = req.get_param("predictor")
        server_domain: Optional[str] = req.get_param("server_domain")

        if keep_types:
            # Limit the number of types to prevent slow SQL queries
            keep_types = keep_types[:10]

        if brands is not None:
            # Limit the number of brands to prevent slow SQL queries
            brands = brands[:10]

        get_insights_ = functools.partial(
            get_insights,
            keep_types=keep_types,
            country=country,
            server_domain=server_domain,
            value_tag=value_tag,
            brands=brands,
            annotated=annotated,
            annotation=annotation,
            barcode=barcode,
            predictor=predictor,
        )

        offset: int = (page - 1) * count
        insights = [i.to_dict() for i in get_insights_(limit=count, offset=offset)]
        response["count"] = get_insights_(count=True)

        if not insights:
            response["insights"] = []
            response["status"] = "no_insights"
        else:
            response["insights"] = insights
            response["status"] = "found"

        resp.media = response


class RandomInsightResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        response: JSONType = {}

        insight_type: Optional[str] = req.get_param("type")
        country: Optional[str] = req.get_param("country")
        value_tag: Optional[str] = req.get_param("value_tag")
        server_domain: Optional[str] = req.get_param("server_domain")
        count: int = req.get_param_as_int("count", min_value=1, default=25)
        predictor = req.get_param("predictor")

        keep_types = [insight_type] if insight_type else None
        get_insights_ = functools.partial(
            get_insights,
            keep_types=keep_types,
            country=country,
            value_tag=value_tag,
            order_by="random",
            server_domain=server_domain,
            predictor=predictor,
        )

        insights = [insight.to_dict() for insight in get_insights_(limit=count)]
        response["count"] = get_insights_(count=True)

        if not insights:
            response["insights"] = []
            response["status"] = "no_insights"
        else:
            response["insights"] = insights
            response["status"] = "found"

        resp.media = response


def parse_auth(req: falcon.Request) -> Optional[OFFAuthentication]:
    session_cookie = req.get_cookie_values("session")

    if session_cookie:
        session_cookie = session_cookie[0]

    authorization = req.get_header("Authorization")

    username = None
    password = None
    if authorization is not None:
        try:
            username, password = basic_decode(authorization)
        except BasicAuthDecodeError:
            raise falcon.HTTPUnauthorized(
                "Invalid authentication, Basic auth expected."
            )

    if not session_cookie and not username and not password:
        return None

    return OFFAuthentication(
        session_cookie=session_cookie, username=username, password=password
    )


def device_id_from_request(req: falcon.Request) -> str:
    """Returns the 'device_id' from the request parameters, or a hash of the
    access route (which should be the IPs of the proxies and the client)."""
    return req.get_param(
        "device_id",
        default=hashlib.sha1(str(req.access_route).encode()).hexdigest(),
    )


class AnnotateInsightResource:
    def on_post(self, req: falcon.Request, resp: falcon.Response):
        insight_id = req.get_param_as_uuid("insight_id", required=True)
        annotation = req.get_param_as_int(
            "annotation", required=True, min_value=-1, max_value=1
        )

        update = req.get_param_as_bool("update", default=True)
        # This field is only needed for nutritional table structure insights.
        data = req.get_param_as_json("data")

        auth = parse_auth(req)
        if auth is not None and auth.get_username() == "null":
            # Smoothie currently sends 'null' as username for anonymous voters
            auth = None

        trusted_annotator = auth is not None

        device_id = device_id_from_request(req)
        logger.info(
            "New annotation received from "
            f"{auth.get_username() if auth else 'unknown annotator'} "
            f"(annotation: {annotation}, insight: {insight_id})"
        )

        annotation_result = save_annotation(
            insight_id,
            annotation,
            update=update,
            data=data,
            auth=auth,
            device_id=device_id,
            trusted_annotator=trusted_annotator,
        )

        resp.media = {
            "status_code": annotation_result.status_code,
            "status": annotation_result.status,
            "description": annotation_result.description,
        }


class IngredientSpellcheckResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        self.spellcheck(req, resp)

    def on_post(self, req: falcon.Request, resp: falcon.Response):
        self.spellcheck(req, resp)

    def spellcheck(self, req: falcon.Request, resp: falcon.Response):
        text = req.get_param("text")
        if text is None:
            barcode = req.get_param("barcode")
            if barcode is None:
                raise falcon.HTTPBadRequest("text or barcode is required.")

            product = get_product(barcode) or {}
            text = product.get("ingredients_text_fr")
            if text is None:
                resp.media = {"status": "not_found"}
                return

        index_name = req.get_param("index", default="product_all")
        confidence = req.get_param_as_float("confidence", default=0.5)
        pipeline = req.get_param_as_list("pipeline") or None
        safe = req.get_param_as_bool("safe", blank_as_true=False)

        if safe is not None and pipeline:
            raise falcon.HTTPBadRequest(
                "pipeline and safe parameters cannot be used together"
            )

        if pipeline:
            for item in pipeline:
                if item not in SPELLCHECKERS:
                    raise falcon.HTTPBadRequest(f"unknown pipeline item: {item}")
        elif safe:
            pipeline = ["patterns", "percentages", "vocabulary"]

        spellchecker = Spellchecker.load(
            client=es_client,
            pipeline=pipeline,
            index_name=index_name,
            confidence=confidence,
        )
        correction_item = spellchecker.correct(text)

        resp.media = {
            "text": text,
            "corrected": correction_item.latest_correction,
            "corrections": correction_item.corrections,
        }


class NutrientPredictorResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        ocr_url = req.get_param("ocr_url", required=True)

        if not ocr_url.endswith(".json"):
            raise falcon.HTTPBadRequest("a JSON file is expected")

        barcode = get_barcode_from_url(ocr_url)

        if barcode is None:
            raise falcon.HTTPBadRequest(f"invalid OCR URL: {ocr_url}")

        try:
            predictions = extract_ocr_predictions(
                barcode, ocr_url, [PredictionType.nutrient]
            )

        except requests.exceptions.RequestException:
            resp.media = {
                "error": "download_error",
                "error_description": "an error occurred during OCR JSON download",
            }
            return

        except OCRParsingException as e:
            logger.error(e)
            resp.media = {
                "error": "invalid_ocr",
                "error_description": "an error occurred during OCR parsing",
            }
            return

        resp.media = {"nutrients": [p.to_dict() for p in predictions]}


class OCRInsightsPredictorResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        ocr_url = req.get_param("ocr_url", required=True)
        barcode = get_barcode_from_url(ocr_url)
        if barcode is None:
            raise falcon.HTTPBadRequest(f"invalid OCR URL: {ocr_url}")

        try:
            insights = extract_ocr_predictions(
                barcode, ocr_url, DEFAULT_OCR_PREDICTION_TYPES
            )

        except requests.exceptions.RequestException:
            resp.media = {
                "error": "download_error",
                "error_description": "an error occurred during OCR JSON download",
            }
            return

        except OCRParsingException as e:
            logger.error(e)
            resp.media = {
                "error": "invalid_ocr",
                "error_description": "an error occurred during OCR parsing",
            }
            return

        resp.media = {
            "insights": insights,
        }


class CategoryPredictorResource:
    @jsonschema.validate(schema.PREDICT_CATEGORY_SCHEMA)
    def on_post(self, req: falcon.Request, resp: falcon.Response):
        """Predict categories using neural categorizer and matching algorithm
        for a specific product."""
        predictors: list[str] = req.media.get("predictors") or ["neural", "matcher"]

        if "barcode" in req.media:
            # Fetch product from DB
            barcode: str = req.media["barcode"]
            product = get_product(barcode) or {}
            if not product:
                raise falcon.HTTPNotFound(description=f"product {barcode} not found")
        else:
            product = req.media["product"]
            if "matcher" in predictors:
                if "lang" not in req.media:
                    raise falcon.HTTPBadRequest(
                        description="lang field is required when using matcher predictor"
                    )
                lang = req.media["lang"]
                product[f"product_name_{lang}"] = product["product_name"]
                product["languages_codes"] = [lang]

        resp.media = predict_category(
            product,
            neural_predictor="neural" in predictors,
            matcher_predictor="matcher" in predictors,
            deepest_only=req.media.get("deepest_only", False),
            threshold=req.media.get("threshold"),
        )


class UpdateDatasetResource:
    def on_post(self, req: falcon.Request, resp: falcon.Response):
        """Re-import the Product Opener product dump."""

        enqueue_job(
            download_product_dataset_job, queue=low_queue, job_kwargs={"timeout": "1h"}
        )

        resp.media = {
            "status": "scheduled",
        }

    def on_get(self, req: falcon.Request, resp: falcon.Response):
        resp.media = {
            "etag": get_product_dataset_etag(),
        }


class ImageImporterResource:
    def on_post(self, req: falcon.Request, resp: falcon.Response):
        barcode = req.get_param("barcode", required=True)
        image_url = req.get_param("image_url", required=True)
        ocr_url = req.get_param("ocr_url", required=True)
        server_domain = req.get_param("server_domain", required=True)

        if server_domain != settings.OFF_SERVER_DOMAIN:
            logger.info("Rejecting image import from %s", server_domain)
            resp.media = {
                "status": "rejected",
            }
            return

        enqueue_job(
            run_import_image_job,
            high_queue,
            job_kwargs={"result_ttl": 0},
            barcode=barcode,
            image_url=image_url,
            ocr_url=ocr_url,
            server_domain=server_domain,
        )
        resp.media = {
            "status": "scheduled",
        }


class ImageCropResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        image_url = req.get_param("image_url", required=True)
        y_min = req.get_param_as_float("y_min", required=True)
        x_min = req.get_param_as_float("x_min", required=True)
        y_max = req.get_param_as_float("y_max", required=True)
        x_max = req.get_param_as_float("x_max", required=True)
        image = get_image_from_url(image_url, session=http_session, error_raise=False)

        if image is None:
            raise falcon.HTTPBadRequest(f"Could not fetch image: {image_url}")

        (left, right, top, bottom) = (
            x_min * image.width,
            x_max * image.width,
            y_min * image.height,
            y_max * image.height,
        )
        cropped_image = image.crop((left, top, right, bottom))
        image_response(cropped_image, resp)


class ImagePredictionImporterResource:
    @jsonschema.validate(schema.IMAGE_PREDICTION_IMPORTER_SCHEMA)
    def on_post(self, req: falcon.Request, resp: falcon.Response):
        timestamp = datetime.datetime.utcnow()
        inserts = []

        for prediction in req.media["predictions"]:
            server_domain: str = prediction.get(
                "server_domain", settings.OFF_SERVER_DOMAIN
            )
            server_type: str = get_server_type(server_domain).name
            source_image = generate_image_path(
                prediction["barcode"], prediction.pop("image_id")
            )
            inserts.append(
                {
                    "timestamp": timestamp,
                    "server_domain": server_domain,
                    "server_type": server_type,
                    "source_image": source_image,
                    **prediction,
                }
            )

        inserted = batch_insert(ImagePrediction, inserts)
        logger.info("{} image predictions inserted".format(inserted))


class ImagePredictionFetchResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        count: int = req.get_param_as_int("count", min_value=1, default=25)
        model_name: Optional[str] = req.get_param("model_name")
        type_: Optional[str] = req.get_param("type")
        model_version: Optional[str] = req.get_param("model_version")
        server_domain: Optional[str] = req.get_param("server_domain")
        barcode: Optional[str] = req.get_param("barcode")
        min_confidence: Optional[float] = req.get_param_as_float("min_confidence")
        random: bool = req.get_param_as_bool("random", default=True)

        where_clauses = []

        if model_name is not None:
            where_clauses.append(ImagePrediction.model_name == model_name)

        if model_version is not None:
            where_clauses.append(ImagePrediction.model_version == model_version)

        if type_ is not None:
            where_clauses.append(ImagePrediction.type == type_)

        if server_domain:
            where_clauses.append(ImageModel.server_domain == server_domain)

        if min_confidence is not None:
            where_clauses.append(ImagePrediction.max_confidence >= min_confidence)

        if barcode is not None:
            where_clauses.append(ImageModel.barcode == barcode)

        query = ImagePrediction.select().join(ImageModel)

        if where_clauses:
            query = query.where(*where_clauses)

        if random:
            query = query.order_by(peewee.fn.Random())

        query = query.limit(count)
        items = [item.to_dict() for item in query.iterator()]
        resp.media = {"predictions": items}


class ImagePredictorResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        image_url = req.get_param("image_url", required=True)
        models: list[str] = req.get_param_as_list("models", required=True)

        available_models = ObjectDetectionModelRegistry.get_available_models()

        for model_name in models:
            if model_name not in available_models:
                raise falcon.HTTPBadRequest(
                    "invalid_model",
                    "unknown model {}, available models: {}"
                    "".format(model_name, ", ".join(available_models)),
                )

        output_image = req.get_param_as_bool("output_image")

        if output_image is None:
            output_image = False

        if output_image and len(models) != 1:
            raise falcon.HTTPBadRequest(
                "invalid_request",
                "a single model must be specified with the `models` parameter "
                "when `output_image` is True",
            )

        image = get_image_from_url(image_url, session=http_session, error_raise=False)

        if image is None:
            raise falcon.HTTPBadRequest(f"Could not fetch image: {image_url}")

        predictions = {}

        for model_name in models:
            model = ObjectDetectionModelRegistry.get(model_name)
            result = model.detect_from_image(image, output_image=output_image)

            if output_image:
                image_response(result.boxed_image, resp)
                return
            else:
                predictions[model_name] = result.to_json()

        resp.media = {"predictions": predictions}


def image_response(image: Image.Image, resp: falcon.Response) -> None:
    resp.content_type = "image/jpeg"
    fp = io.BytesIO()
    image.save(fp, "JPEG")
    resp.stream_len = fp.tell()
    fp.seek(0)
    resp.stream = fp


class ImageLogoResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        logo_ids: list[str] = req.get_param_as_list(
            "logo_ids", transform=int, required=True
        )
        logos = []
        for logo in (
            LogoAnnotation.select()
            .join(ImagePrediction)
            .join(ImageModel)
            .where(LogoAnnotation.id.in_(logo_ids))
            .iterator()
        ):
            logo_dict = logo.to_dict()
            image_prediction = logo_dict.pop("image_prediction")
            logo_dict["image"] = image_prediction["image"]
            logos.append(logo_dict)

        resp.media = {"logos": logos, "count": len(logos)}


class ImageLogoSearchResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        count: int = req.get_param_as_int(
            "count", min_value=1, max_value=2000, default=25
        )
        type_: Optional[str] = req.get_param("type")
        barcode: Optional[str] = req.get_param("barcode")
        value: Optional[str] = req.get_param("value")
        taxonomy_value: Optional[str] = req.get_param("taxonomy_value")
        min_confidence: Optional[float] = req.get_param_as_float("min_confidence")
        random: bool = req.get_param_as_bool("random", default=False)
        server_domain: Optional[str] = req.get_param("server_domain")
        annotated: Optional[bool] = req.get_param_as_bool("annotated")

        if type_ is None and (value is not None or taxonomy_value is not None):
            raise falcon.HTTPBadRequest(
                description="a type is required if `value` or `taxonomy_value` is provided"
            )

        if value is not None and taxonomy_value is not None:
            raise falcon.HTTPBadRequest(
                description="`value` and `taxonomy_value` are mutually exclusive parameters"
            )

        if type_ == "label" and taxonomy_value is None and value is not None:
            raise falcon.HTTPBadRequest(
                description="you should provide a `taxonomy_value` and not a `value` "
                "for label type"
            )

        where_clauses = []
        if annotated is not None:
            where_clauses.append(LogoAnnotation.annotation_value.is_null(not annotated))

        if min_confidence is not None:
            where_clauses.append(LogoAnnotation.score >= min_confidence)

        join_image_prediction = False
        join_image_model = False

        if server_domain:
            where_clauses.append(ImageModel.server_domain == server_domain)
            join_image_model = True

        if barcode is not None:
            where_clauses.append(ImageModel.barcode == barcode)
            join_image_model = True

        if type_ is not None:
            where_clauses.append(LogoAnnotation.annotation_type == type_)

        if value is not None:
            value_tag = get_tag(value)
            where_clauses.append(LogoAnnotation.annotation_value_tag == value_tag)

        if taxonomy_value is not None:
            where_clauses.append(LogoAnnotation.taxonomy_value == taxonomy_value)

        query = LogoAnnotation.select()
        join_image_prediction = join_image_prediction or join_image_model

        if join_image_prediction:
            query = query.join(ImagePrediction)

            if join_image_model:
                query = query.join(ImageModel)

        if where_clauses:
            query = query.where(*where_clauses)

        query_count = query.count()

        if random:
            query = query.order_by(peewee.fn.Random())

        query = query.limit(count)
        items = [item.to_dict() for item in query.iterator()]

        for item in items:
            image_prediction = item.pop("image_prediction")
            item["image"] = image_prediction["image"]

        resp.media = {"logos": items, "count": query_count}


class ImageLogoDetailResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response, logo_id: int):
        logo = LogoAnnotation.get_or_none(id=logo_id)

        if logo is None:
            resp.status = falcon.HTTP_404
            return

        logo_dict = logo.to_dict()
        image_prediction = logo_dict.pop("image_prediction")
        logo_dict["image"] = image_prediction["image"]
        resp.media = logo_dict

    @jsonschema.validate(schema.UPDATE_LOGO_SCHEMA)
    def on_put(self, req: falcon.Request, resp: falcon.Response, logo_id: int):
        logo = LogoAnnotation.get_or_none(id=logo_id)

        if logo is None:
            resp.status = falcon.HTTP_404
            return

        type_ = req.media["type"]
        value = req.media["value"] or None
        updated = False

        if type_ != logo.annotation_type:
            logo.annotation_type = type_
            updated = True

        if value != logo.annotation_value:
            logo.annotation_value = value

            if value is not None:
                value_tag = get_tag(value)
                logo.annotation_value_tag = value_tag
                logo.taxonomy_value = match_taxonomized_value(value_tag, type_)
            else:
                logo.annotation_value_tag = None
                logo.taxonomy_value = None

            updated = True

        if updated:
            auth = parse_auth(req)
            username = None if auth is None else auth.get_username()
            logo.username = username
            logo.completed_at = datetime.datetime.utcnow()
            logo.save()

        resp.status = falcon.HTTP_204


class ImageLogoAnnotateResource:
    @jsonschema.validate(schema.ANNOTATE_LOGO_SCHEMA)
    def on_post(self, req: falcon.Request, resp: falcon.Response):
        server_domain = req.media.get("server_domain", settings.OFF_SERVER_DOMAIN)
        annotations = req.media["annotations"]
        auth = parse_auth(req)
        username = None if auth is None else auth.get_username()
        completed_at = datetime.datetime.utcnow()
        annotated_logos = []

        for annotation in annotations:
            logo_id = annotation["logo_id"]
            type_ = annotation["type"]
            value = annotation["value"] or None
            try:
                logo = LogoAnnotation.get_by_id(logo_id)
            except LogoAnnotation.DoesNotExist:
                raise falcon.HTTPNotFound(description=f"logo {logo_id} not found")

            if logo.annotation_type is not None:
                # Logo is already annotated, skip
                continue

            if value is not None:
                if type_ == "label" and not is_prefixed_value(value):
                    raise falcon.HTTPBadRequest(
                        description=f"language-prefixed value are required for label type (here: {value})"
                    )
                logo.annotation_value = value
                value_tag = get_tag(value)
                logo.annotation_value_tag = value_tag
                logo.taxonomy_value = match_taxonomized_value(value_tag, type_)
            elif type_ in ("brand", "category", "label", "store"):
                raise falcon.HTTPBadRequest(
                    description=f"value required for type {type_} (logo {logo_id})"
                )

            logo.annotation_type = type_
            logo.username = username
            logo.completed_at = completed_at
            annotated_logos.append(logo)

        for logo in annotated_logos:
            logo.save()

        created = generate_insights_from_annotated_logos(annotated_logos, server_domain)
        resp.media = {"created insights": created}


class ImageLogoUpdateResource:
    def on_post(self, req: falcon.Request, resp: falcon.Response):
        """Bulk update logo annotations: change type and value of logos that have specific
        types and values.

        Because this endpoint mass-update annotations, leave it out of API documentation."""
        source_value = req.get_param("source_value", required=True)
        source_type = req.get_param("source_type", required=True)
        target_value = req.get_param("target_value", required=True)
        target_type = req.get_param("target_type", required=True)

        auth = parse_auth(req)
        username = None if auth is None else auth.get_username()
        completed_at = datetime.datetime.utcnow()

        target_value_tag = get_tag(target_value)
        source_value_tag = get_tag(source_value)
        taxonomy_value = match_taxonomized_value(target_value_tag, target_type)

        query = LogoAnnotation.update(
            {
                LogoAnnotation.annotation_type: target_type,
                LogoAnnotation.annotation_value: target_value,
                LogoAnnotation.annotation_value_tag: target_value_tag,
                LogoAnnotation.taxonomy_value: taxonomy_value,
                LogoAnnotation.username: username,
                LogoAnnotation.completed_at: completed_at,
            }
        ).where(
            LogoAnnotation.annotation_type == source_type,
            LogoAnnotation.annotation_value_tag == source_value_tag,
        )
        updated = query.execute()
        resp.media = {"updated": updated}


class WebhookProductResource:
    """This handles requests from product opener
    that act as webhooks on product update or deletion.
    """

    def on_post(self, req: falcon.Request, resp: falcon.Response):
        barcode = req.get_param("barcode", required=True)
        action = req.get_param("action", required=True)
        server_domain = req.get_param("server_domain", required=True)
        if server_domain != settings.OFF_SERVER_DOMAIN:
            logger.info("Rejecting webhook event from {}".format(server_domain))
            resp.media = {
                "status": "rejected",
            }
            return

        logger.info(
            "New webhook event received for product {} (action: {}, "
            "domain: {})".format(barcode, action, server_domain)
        )
        if action not in ("updated", "deleted"):
            raise falcon.HTTPBadRequest(
                title="invalid_action",
                description="action must be one of " "`deleted`, `updated`",
            )

        if action == "updated":
            enqueue_in_job(
                update_insights_job,
                high_queue,
                settings.UPDATED_PRODUCT_WAIT,
                barcode=barcode,
                server_domain=server_domain,
            )
        elif action == "deleted":
            enqueue_job(
                delete_product_insights_job,
                high_queue,
                job_kwargs={"result_ttl": 0},
                barcode=barcode,
                server_domain=server_domain,
            )

        resp.media = {
            "status": "scheduled",
        }


class ProductQuestionsResource:
    """Get questions about a product to confirm/infirm an insight

    see also doc/explanation/questions.md
    """

    def on_get(self, req: falcon.Request, resp: falcon.Response, barcode: str):
        response: JSONType = {}
        count: int = req.get_param_as_int("count", min_value=1, default=25)
        lang: str = req.get_param("lang", default="en")
        # If the device_id is not provided as a request parameter, we use the
        # hash of the IPs as a backup.
        device_id = device_id_from_request(req)
        server_domain: Optional[str] = req.get_param("server_domain")

        auth: Optional[OFFAuthentication] = parse_auth(req)

        keep_types = QuestionFormatterFactory.get_default_types()

        insights = list(
            get_insights(
                barcode=barcode,
                keep_types=keep_types,
                server_domain=server_domain,
                limit=count,
                order_by="n_votes",
                avoid_voted_on=_get_skip_voted_on(auth, device_id),
                automatically_processable=False,
            )
        )

        if not insights:
            response["questions"] = []
            response["status"] = "no_questions"
        else:
            questions: list[JSONType] = []

            for insight in insights:
                formatter_cls = QuestionFormatterFactory.get(insight.type)
                formatter: QuestionFormatter = formatter_cls(TRANSLATION_STORE)
                question = formatter.format_question(insight, lang)
                questions.append(question.serialize())

            response["questions"] = questions
            response["status"] = "found"

        resp.media = response


class RandomQuestionsResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        get_questions_resource_on_get(req, resp, "random")


class PopularQuestionsResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        get_questions_resource_on_get(req, resp, "popularity")


def get_questions_resource_on_get(
    req: falcon.Request, resp: falcon.Response, order_by: str
):
    response: JSONType = {}
    page: int = req.get_param_as_int("page", min_value=1, default=1)
    count: int = req.get_param_as_int("count", min_value=1, default=25)
    lang: str = req.get_param("lang", default="en")
    keep_types: Optional[list[str]] = req.get_param_as_list(
        "insight_types", required=False
    )
    country: Optional[str] = req.get_param("country")
    value_tag: str = req.get_param("value_tag")
    brands = req.get_param_as_list("brands") or None
    server_domain: Optional[str] = req.get_param("server_domain")
    reserved_barcode: Optional[bool] = req.get_param_as_bool(
        "reserved_barcode", default=False
    )
    # filter by annotation campaign
    campaign: Optional[str] = req.get_param("campaign")
    predictor = req.get_param("predictor")

    # If the device_id is not provided as a request parameter, we use the
    # hash of the IPs as a backup.
    device_id = device_id_from_request(req)

    auth: Optional[OFFAuthentication] = parse_auth(req)

    if reserved_barcode:
        # Include all results, including non reserved barcodes
        reserved_barcode = None

    if keep_types is None:
        keep_types = QuestionFormatterFactory.get_default_types()
    else:
        # Limit the number of types to prevent slow SQL queries
        keep_types = keep_types[:10]

    if brands is not None:
        # Limit the number of brands to prevent slow SQL queries
        brands = brands[:10]

    get_insights_ = functools.partial(
        get_insights,
        keep_types=keep_types,
        country=country,
        server_domain=server_domain,
        value_tag=value_tag,
        brands=brands,
        order_by=order_by,
        reserved_barcode=reserved_barcode,
        avoid_voted_on=_get_skip_voted_on(auth, device_id),
        automatically_processable=False,
        campaign=campaign,
        predictor=predictor,
    )

    offset: int = (page - 1) * count
    insights = list(get_insights_(limit=count, offset=offset))
    response["count"] = get_insights_(count=True)
    # This code should be merged with the one in ProductQuestionsResource.get
    if not insights:
        response["questions"] = []
        response["status"] = "no_questions"
    else:
        questions: list[JSONType] = []

        for insight in insights:
            formatter_cls = QuestionFormatterFactory.get(insight.type)

            if formatter_cls is None:
                continue

            formatter: QuestionFormatter = formatter_cls(TRANSLATION_STORE)
            question = formatter.format_question(insight, lang)
            questions.append(question.serialize())

        response["questions"] = questions
        response["status"] = "found"

    resp.media = response


class StatusResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        resp.media = {
            "status": "running",
        }


class HealthResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        from robotoff.health import health

        message, status, headers = health.run()
        resp.media = {
            "message": orjson.loads(message),
            "status": status,
            "headers": headers,
        }
        resp.status = str(status)


class DumpResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        keep_types: list[str] = req.get_param_as_list(
            "insight_types", required=False, default=[]
        )[:10]

        barcode = req.get_param("barcode")
        annotated = req.get_param_as_bool("annotated", blank_as_true=False)
        value_tag = req.get_param("value_tag")

        insights_iter = get_insights(
            barcode=barcode,
            keep_types=keep_types,
            annotated=annotated,
            value_tag=value_tag,
            limit=None,
        )

        writer = None

        with tempfile.TemporaryFile("w+", newline="") as temp_f:
            for insight in insights_iter:
                serial = orjson.loads(orjson.dumps(insight.to_dict()))

                if writer is None:
                    writer = csv.DictWriter(temp_f, fieldnames=serial.keys())
                    writer.writeheader()

                writer.writerow(serial)

            temp_f.seek(0)
            content = temp_f.read()

        if content:
            resp.content_type = "text/csv"
            resp.body = content
        else:
            resp.status = falcon.HTTP_204


class UserStatisticsResource:
    def on_get(self, req: falcon.Request, resp: falcon.Response, username: str):
        annotation_count = (
            ProductInsight.select().where(ProductInsight.username == username).count()
        )
        resp.media = {"count": {"annotations": annotation_count}}


class ImageCollection:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        response: JSONType = {}
        count: int = req.get_param_as_int("count", min_value=1, default=25)
        page: int = req.get_param_as_int("page", min_value=1, default=1)
        with_predictions: Optional[bool] = req.get_param_as_bool(
            "with_predictions", default=False
        )
        barcode: Optional[str] = req.get_param("barcode")
        server_domain = settings.OFF_SERVER_DOMAIN

        get_images_ = functools.partial(
            get_images,
            with_predictions=with_predictions,
            barcode=barcode,
            server_domain=server_domain,
        )

        offset: int = (page - 1) * count
        images = [i.to_dict() for i in get_images_(limit=count, offset=offset)]
        response["count"] = get_images_(count=True)

        if not images:
            response["images"] = []
            response["status"] = "no_images"
        else:
            response["images"] = images
            response["status"] = "found"

        resp.media = response


class PredictionCollection:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        response: JSONType = {}
        page: int = req.get_param_as_int("page", min_value=1, default=1)
        count: int = req.get_param_as_int("count", min_value=1, default=25)
        barcode: Optional[str] = req.get_param("barcode")
        value_tag: str = req.get_param("value_tag")
        keep_types: Optional[list[str]] = req.get_param_as_list(
            "insight_types", required=False
        )
        brands = req.get_param_as_list("brands") or None
        server_domain: Optional[str] = req.get_param("server_domain")

        if keep_types:
            # Limit the number of types to prevent slow SQL queries
            keep_types = keep_types[:10]

        if brands is not None:
            # Limit the number of brands to prevent slow SQL queries
            brands = brands[:10]

        query_parameters = {
            "server_domain": server_domain,
            "keep_types": keep_types,
            "value_tag": value_tag,
            "barcode": barcode,
        }

        get_predictions_ = functools.partial(get_predictions, **query_parameters)

        offset: int = (page - 1) * count
        predictions = [
            i.to_dict() for i in get_predictions_(limit=count, offset=offset)
        ]

        response["count"] = get_predictions_(count=True)

        if not predictions:
            response["predictions"] = []
            response["status"] = "no_predictions"
        else:
            response["predictions"] = list(predictions)
            response["status"] = "found"

        resp.media = response


class UnansweredQuestionCollection:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        response: JSONType = {}
        page: int = req.get_param_as_int("page", min_value=1, default=1)
        count: int = req.get_param_as_int("count", min_value=1, default=25)
        insight_type: str = req.get_param("type")
        country: Optional[str] = req.get_param("country")
        server_domain: Optional[str] = req.get_param("server_domain")
        reserved_barcode: Optional[bool] = req.get_param_as_bool(
            "reserved_barcode", default=False
        )
        campaign: Optional[str] = req.get_param("campaign")
        predictor = req.get_param("predictor")

        get_insights_ = functools.partial(
            get_insights,
            keep_types=[insight_type] if insight_type else None,
            group_by_value_tag=True,
            limit=count,
            country=country,
            server_domain=server_domain,
            automatically_processable=False,
            reserved_barcode=reserved_barcode,
            campaign=campaign,
            predictor=predictor,
        )

        offset: int = (page - 1) * count
        insights = [i for i in get_insights_(limit=count, offset=offset)]

        response["count"] = get_insights_(count=True)

        if not insights:
            response["questions"] = []
            response["status"] = "no_questions"
        else:
            response["questions"] = insights
            response["status"] = "found"

        resp.media = response


class ImagePredictionCollection:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        response: JSONType = {}
        count: int = req.get_param_as_int("count", min_value=1, default=25)
        page: int = req.get_param_as_int("page", min_value=1, default=1)
        with_logo: Optional[bool] = req.get_param_as_bool("with_logo", default=False)
        barcode: Optional[str] = req.get_param("barcode")
        type: Optional[str] = req.get_param("type")
        server_domain: Optional[str] = req.get_param("server_domain")

        query_parameters = {
            "with_logo": with_logo,
            "barcode": barcode,
            "type": type,
            "server_domain": server_domain,
        }

        get_image_predictions_ = functools.partial(
            get_image_predictions, **query_parameters
        )

        offset: int = (page - 1) * count
        images = [
            i.to_dict() for i in get_image_predictions_(limit=count, offset=offset)
        ]
        response["count"] = get_image_predictions_(count=True)

        if not images:
            response["image_predictions"] = []
            response["status"] = "no_image_predictions"
        else:
            response["image_predictions"] = images
            response["status"] = "found"

        resp.media = response


class LogoAnnotationCollection:
    def on_get(self, req: falcon.Request, resp: falcon.Response):
        response: JSONType = {}
        barcode: Optional[str] = req.get_param("barcode")
        keep_types: Optional[list[str]] = req.get_param_as_list("types", required=False)
        value_tag: str = req.get_param("value_tag")
        page: int = req.get_param_as_int("page", min_value=1, default=1)
        count: int = req.get_param_as_int("count", min_value=1, default=25)
        server_domain: Optional[str] = req.get_param("server_domain")

        if keep_types:
            # Limit the number of types to prevent slow SQL queries
            keep_types = keep_types[:10]

        query_parameters = {
            "server_domain": server_domain,
            "barcode": barcode,
            "keep_types": keep_types,
            "value_tag": value_tag,
        }

        get_annotation_ = functools.partial(get_logo_annotation, **query_parameters)

        offset: int = (page - 1) * count
        annotation = [i.to_dict() for i in get_annotation_(limit=count, offset=offset)]
        response["count"] = get_annotation_(count=True)

        if not annotation:
            response["annotation"] = []
            response["status"] = "no_annotation"
        else:
            response["annotation"] = annotation
            response["status"] = "found"

        resp.media = response


cors = CORS(
    allow_all_origins=True,
    allow_all_headers=True,
    allow_all_methods=True,
    allow_credentials_all_origins=True,
    max_age=600,
)

api = falcon.API(
    middleware=[cors.middleware, MultipartMiddleware(), DBConnectionMiddleware()]
)

json_handler = falcon.media.JSONHandler(dumps=orjson.dumps, loads=orjson.loads)
extra_handlers = {
    "application/json": json_handler,
}

api.resp_options.media_handlers.update(extra_handlers)

# Parse form parameters
api.req_options.auto_parse_form_urlencoded = True
api.req_options.strip_url_path_trailing_slash = True
api.req_options.auto_parse_qs_csv = True
# defines urls
api.add_route("/api/v1/insights/{barcode}", ProductInsightResource())
api.add_route("/api/v1/insights/detail/{insight_id:uuid}", ProductInsightDetail())
api.add_route("/api/v1/insights", InsightCollection())
api.add_route("/api/v1/insights/random", RandomInsightResource())
api.add_route("/api/v1/insights/annotate", AnnotateInsightResource())
api.add_route("/api/v1/predict/ingredients/spellcheck", IngredientSpellcheckResource())
api.add_route("/api/v1/predict/nutrient", NutrientPredictorResource())
api.add_route("/api/v1/predict/ocr_insights", OCRInsightsPredictorResource())
api.add_route("/api/v1/predict/category", CategoryPredictorResource())
api.add_route("/api/v1/products/dataset", UpdateDatasetResource())
api.add_route("/api/v1/webhook/product", WebhookProductResource())
api.add_route("/api/v1/images/import", ImageImporterResource())
api.add_route("/api/v1/images/crop", ImageCropResource())
api.add_route("/api/v1/images/predictions/import", ImagePredictionImporterResource())
api.add_route("/api/v1/images/predictions", ImagePredictionFetchResource())
api.add_route("/api/v1/images/predict", ImagePredictorResource())
api.add_route("/api/v1/images/logos", ImageLogoResource())
api.add_route("/api/v1/images/logos/search", ImageLogoSearchResource())
api.add_route("/api/v1/images/logos/{logo_id:int}", ImageLogoDetailResource())
api.add_route("/api/v1/images/logos/annotate", ImageLogoAnnotateResource())
api.add_route("/api/v1/images/logos/update", ImageLogoUpdateResource())
api.add_route("/api/v1/questions/{barcode}", ProductQuestionsResource())
api.add_route("/api/v1/questions/random", RandomQuestionsResource())
api.add_route("/api/v1/questions/popular", PopularQuestionsResource())
api.add_route("/api/v1/questions/unanswered", UnansweredQuestionCollection())
api.add_route("/api/v1/status", StatusResource())
api.add_route("/api/v1/health", HealthResource())
api.add_route("/api/v1/dump", DumpResource())
api.add_route("/api/v1/users/statistics/{username}", UserStatisticsResource())
api.add_route("/api/v1/predictions", PredictionCollection())
api.add_route("/api/v1/images/prediction/collection", ImagePredictionCollection())
api.add_route("/api/v1/images", ImageCollection())
api.add_route("/api/v1/annotation/collection", LogoAnnotationCollection())



import enum
import math
import operator
import re
from collections import Counter, defaultdict
from typing import Callable, Optional, Union

from robotoff.utils import get_logger
from robotoff.utils.types import JSONType

MULTIPLE_SPACES_REGEX = re.compile(r" {2,}")

logger = get_logger(__name__)


class OCRParsingException(Exception):
    pass


class OCRField(enum.Enum):
    full_text = 1
    full_text_contiguous = 2
    text_annotations = 3


class OCRRegex:
    __slots__ = ("regex", "field", "lowercase", "processing_func", "priority", "notify")

    def __init__(
        self,
        regex: re.Pattern,
        field: OCRField,
        lowercase: bool = False,
        processing_func: Optional[Callable] = None,
        priority: Optional[int] = None,
        notify: bool = False,
    ):
        self.regex: re.Pattern = regex
        self.field: OCRField = field
        self.lowercase: bool = lowercase
        self.processing_func: Optional[Callable] = processing_func
        self.priority = priority
        self.notify = notify


class ImageOrientation(enum.Enum):
    up = 1  # intended orientation
    down = 2  # 180° rotation
    left = 3  # 90° counterclockwise rotation
    right = 4  # 90° clockwise rotation
    unknown = 5


class OrientationResult:
    __slots__ = ("count", "orientation")

    def __init__(self, count: Counter):
        most_common_list = count.most_common(1)
        self.orientation: ImageOrientation

        if most_common_list:
            self.orientation = most_common_list[0][0]
        else:
            self.orientation = ImageOrientation.unknown

        self.count: dict[str, int] = {key.name: value for key, value in count.items()}

    def to_json(self) -> JSONType:
        return {
            "count": self.count,
            "orientation": self.orientation.name,
        }


class OCRResultGenerationException(Exception):
    """An Error occurred while analyzing OCR

    args may contain ocr_url"""

    pass


class OCRResult:
    __slots__ = (
        "text_annotations",
        "text_annotations_str",
        "text_annotations_str_lower",
        "full_text_annotation",
        "logo_annotations",
        "safe_search_annotation",
        "label_annotations",
    )

    def __init__(self, data: JSONType, lazy: bool = True):
        self.text_annotations: list[OCRTextAnnotation] = []
        self.full_text_annotation: Optional[OCRFullTextAnnotation] = None
        self.logo_annotations: list[LogoAnnotation] = []
        self.label_annotations: list[LabelAnnotation] = []
        self.safe_search_annotation: Optional[SafeSearchAnnotation] = None

        for text_annotation_data in data.get("textAnnotations", []):
            text_annotation = OCRTextAnnotation(text_annotation_data)
            self.text_annotations.append(text_annotation)

        self.text_annotations_str: str = ""
        self.text_annotations_str_lower: str = ""

        if self.text_annotations:
            self.text_annotations_str = self.text_annotations[0].text
            self.text_annotations_str_lower = self.text_annotations_str.lower()

        full_text_annotation_data = data.get("fullTextAnnotation")

        if full_text_annotation_data:
            self.full_text_annotation = OCRFullTextAnnotation(
                full_text_annotation_data, lazy=lazy
            )

        for logo_annotation_data in data.get("logoAnnotations", []):
            logo_annotation = LogoAnnotation(logo_annotation_data)
            self.logo_annotations.append(logo_annotation)

        for label_annotation_data in data.get("labelAnnotations", []):
            label_annotation = LabelAnnotation(label_annotation_data)
            self.label_annotations.append(label_annotation)

        if "safeSearchAnnotation" in data:
            self.safe_search_annotation = SafeSearchAnnotation(
                data["safeSearchAnnotation"]
            )

    def get_full_text(self, lowercase: bool = False) -> str:
        if self.full_text_annotation is not None:
            if lowercase:
                return self.full_text_annotation.text_lower

            return self.full_text_annotation.text

        return ""

    def get_full_text_contiguous(self, lowercase: bool = False) -> str:
        if self.full_text_annotation is not None:
            if lowercase:
                return self.full_text_annotation.contiguous_text_lower

            return self.full_text_annotation.contiguous_text

        return ""

    def get_text_annotations(self, lowercase: bool = False) -> str:
        if lowercase:
            return self.text_annotations_str_lower
        else:
            return self.text_annotations_str

    def _get_text(self, field: OCRField, lowercase: bool) -> str:
        if field == OCRField.full_text:
            text = self.get_full_text(lowercase)

            if text is None:
                # If there is no full text, get text annotations as fallback
                return self.get_text_annotations(lowercase)
            else:
                return text

        elif field == OCRField.full_text_contiguous:
            text = self.get_full_text_contiguous(lowercase)

            if text is None:
                # If there is no full text, get text annotations as fallback
                return self.get_text_annotations(lowercase)
            else:
                return text

        elif field == OCRField.text_annotations:
            return self.get_text_annotations(lowercase)

        else:
            raise ValueError("invalid field: {}".format(field))

    def get_text(self, ocr_regex: OCRRegex) -> str:
        return self._get_text(ocr_regex.field, ocr_regex.lowercase)

    def get_logo_annotations(self) -> list["LogoAnnotation"]:
        return self.logo_annotations

    def get_label_annotations(self) -> list["LabelAnnotation"]:
        return self.label_annotations

    def get_safe_search_annotation(self):
        return self.safe_search_annotation

    def get_orientation(self) -> Optional[OrientationResult]:
        if self.full_text_annotation:
            return self.full_text_annotation.detect_orientation()
        else:
            return None

    @classmethod
    def from_json(cls, data: JSONType, **kwargs) -> Optional["OCRResult"]:
        if "responses" not in data or not isinstance(data["responses"], list):
            raise OCRParsingException("Responses field (list) expected in OCR JSON")

        responses = data["responses"]

        if not responses:
            raise OCRParsingException("Empty OCR response")

        response = responses[0]
        if "error" in response:
            raise OCRParsingException(f"Error in OCR response: {response['error']}")

        try:
            return OCRResult(response, **kwargs)
        except Exception as e:
            raise OCRParsingException("Error during OCR parsing") from e

    def get_languages(self) -> Optional[dict[str, int]]:
        if self.full_text_annotation is not None:
            return self.full_text_annotation.get_languages()

        return None


def get_text(
    content: Union[OCRResult, str],
    ocr_regex: Optional[OCRRegex] = None,
    lowercase: bool = True,
) -> str:
    if isinstance(content, str):
        if ocr_regex and ocr_regex.lowercase:
            return content.lower()

        return content.lower() if lowercase else content

    elif isinstance(content, OCRResult):
        if ocr_regex:
            return content.get_text(ocr_regex)
        else:
            text = content.get_full_text_contiguous(lowercase=lowercase)

            if not text:
                text = content.get_text_annotations(lowercase=lowercase)

            return text

    raise TypeError("invalid type: {}".format(type(content)))


class OCRFullTextAnnotation:
    __slots__ = (
        "text",
        "text_lower",
        "_pages",
        "_pages_data",
        "contiguous_text",
        "contiguous_text_lower",
    )

    def __init__(self, data: JSONType, lazy: bool = True):
        self.text = MULTIPLE_SPACES_REGEX.sub(" ", data["text"])
        self.text_lower = self.text.lower()
        self.contiguous_text = self.text.replace("\n", " ")
        self.contiguous_text = MULTIPLE_SPACES_REGEX.sub(" ", self.contiguous_text)
        self.contiguous_text_lower = self.contiguous_text.lower()
        self._pages_data = data["pages"]
        self._pages: list[TextAnnotationPage] = []

        if not lazy:
            self.load_pages()

    def get_languages(self) -> dict[str, int]:
        counts: dict[str, int] = defaultdict(int)
        for page in self.pages:
            page_counts = page.get_languages()

            for key, value in page_counts.items():
                counts[key] += value

        return dict(counts)

    @property
    def pages(self) -> list["TextAnnotationPage"]:
        if self._pages_data is not None:
            self.load_pages()

        return self._pages

    def load_pages(self):
        self._pages = [TextAnnotationPage(page) for page in self._pages_data]
        self._pages_data = None

    def detect_orientation(self) -> OrientationResult:
        word_orientations: list[ImageOrientation] = []

        for page in self.pages:
            word_orientations += page.detect_words_orientation()

        count = Counter(word_orientations)
        return OrientationResult(count)


class TextAnnotationPage:
    def __init__(self, data: JSONType):
        self.width = data["width"]
        self.height = data["height"]
        self.blocks: list[Block] = [Block(d) for d in data["blocks"]]

    def get_languages(self) -> dict[str, int]:
        counts: dict[str, int] = defaultdict(int)
        for block in self.blocks:
            block_counts = block.get_languages()

            for key, value in block_counts.items():
                counts[key] += value

        return dict(counts)

    def detect_words_orientation(self) -> list[ImageOrientation]:
        word_orientations: list[ImageOrientation] = []

        for block in self.blocks:
            word_orientations += block.detect_words_orientation()

        return word_orientations


class Block:
    def __init__(self, data: JSONType):
        self.type = data["blockType"]
        self.paragraphs: list[Paragraph] = [
            Paragraph(paragraph) for paragraph in data["paragraphs"]
        ]

        self.bounding_poly = None
        if "boundingBox" in data:
            self.bounding_poly = BoundingPoly(data["boundingBox"])

    def get_languages(self) -> dict[str, int]:
        counts: dict[str, int] = defaultdict(int)
        for paragraph in self.paragraphs:
            paragraph_counts = paragraph.get_languages()

            for key, value in paragraph_counts.items():
                counts[key] += value

        return dict(counts)

    def detect_orientation(self) -> Optional[ImageOrientation]:
        if self.bounding_poly:
            return self.bounding_poly.detect_orientation()

        return None

    def detect_words_orientation(self) -> list[ImageOrientation]:
        word_orientations: list[ImageOrientation] = []

        for paragraph in self.paragraphs:
            word_orientations += paragraph.detect_words_orientation()

        return word_orientations


class Paragraph:
    def __init__(self, data: JSONType):
        self.words: list[Word] = [Word(word) for word in data["words"]]

        self.bounding_poly = None
        if "boundingBox" in data:
            self.bounding_poly = BoundingPoly(data["boundingBox"])

    def get_languages(self) -> dict[str, int]:
        counts: dict[str, int] = defaultdict(int)

        for word in self.words:
            if word.languages is not None:
                for language in word.languages:
                    counts[language.language] += 1
            else:
                counts["null"] += 1

        counts["words"] = len(self.words)
        return dict(counts)

    def detect_orientation(self) -> Optional[ImageOrientation]:
        if self.bounding_poly:
            return self.bounding_poly.detect_orientation()

        return None

    def detect_words_orientation(self) -> list[ImageOrientation]:
        return [word.detect_orientation() for word in self.words]

    def get_text(self) -> str:
        """Return the text of the paragraph, by concatenating the words."""
        return "".join(w.get_text() for w in self.words)


class Word:
    __slots__ = ("bounding_poly", "symbols", "languages")

    def __init__(self, data: JSONType):
        self.bounding_poly = BoundingPoly(data["boundingBox"])
        self.symbols: list[Symbol] = [Symbol(s) for s in data["symbols"]]

        self.languages: Optional[list[DetectedLanguage]] = None
        word_property = data.get("property", {})

        if "detectedLanguages" in word_property:
            self.languages = [
                DetectedLanguage(lang) for lang in data["property"]["detectedLanguages"]
            ]

    def get_text(self) -> str:
        text_list = []
        for symbol in self.symbols:
            symbol_str = ""

            if symbol.symbol_break and symbol.symbol_break.is_prefix:
                symbol_str = symbol.symbol_break.get_value()

            symbol_str += symbol.text

            if symbol.symbol_break and not symbol.symbol_break.is_prefix:
                symbol_str += symbol.symbol_break.get_value()

            text_list.append(symbol_str)

        return "".join(text_list)

    def detect_orientation(self) -> ImageOrientation:
        return self.bounding_poly.detect_orientation()

    def on_same_line(self, word: "Word"):
        (
            self_alpha,
            self_width,
        ) = self.bounding_poly.get_direction_vector_alpha_distance()
        (
            word_alpha,
            word_width,
        ) = word.bounding_poly.get_direction_vector_alpha_distance()
        self_symbol_width = self_width / len(self.symbols)
        word_symbol_width = word_width / len(word.symbols)
        print(
            "Alpha/distance/mean symbol width:",
            self_alpha,
            self_width,
            self_symbol_width,
        )
        print(
            "Alpha/distance/mean symbol width:",
            word_alpha,
            word_width,
            word_symbol_width,
        )


class Symbol:
    __slots__ = ("bounding_poly", "text", "confidence", "symbol_break")

    def __init__(self, data: JSONType):
        self.bounding_poly: Optional[BoundingPoly] = None
        if "boundingBox" in data:
            self.bounding_poly = BoundingPoly(data["boundingBox"])

        self.text = data["text"]
        self.confidence = data.get("confidence", None)

        self.symbol_break: Optional[DetectedBreak] = None
        symbol_property = data.get("property", {})

        if "detectedBreak" in symbol_property:
            self.symbol_break = DetectedBreak(symbol_property["detectedBreak"])

    def detect_orientation(self) -> Optional[ImageOrientation]:
        if self.bounding_poly:
            return self.bounding_poly.detect_orientation()
        return None


class DetectedBreak:
    __slots__ = ("type", "is_prefix")

    def __init__(self, data: JSONType):
        self.type = data["type"]
        self.is_prefix = data.get("isPrefix", False)

    def __repr__(self):
        return "<DetectedBreak {}>".format(self.type)

    def get_value(self):
        if self.type in ("UNKNOWN", "HYPHEN"):
            return ""

        elif self.type in ("SPACE", "SURE_SPACE", "EOL_SURE_SPACE"):
            return " "

        elif self.type == "LINE_BREAK":
            return "\n"

        else:
            raise ValueError("unknown type: {}".format(self.type))


class DetectedLanguage:
    __slots__ = ("language", "confidence")

    def __init__(self, data: JSONType):
        self.language = data["languageCode"]
        self.confidence = data.get("confidence", 0)

    def __repr__(self):
        return "<DetectedLanguage: {} (confidence: {})>".format(
            self.language, self.confidence
        )


class BoundingPoly:
    __slots__ = ("vertices",)

    def __init__(self, data: JSONType):
        self.vertices = [
            (point.get("x", 0), point.get("y", 0)) for point in data["vertices"]
        ]

    def get_direction_vector(self) -> list[tuple[int, int]]:
        left_point = (
            (self.vertices[0][0] + self.vertices[3][0]) / 2,
            (self.vertices[0][1] + self.vertices[3][1]) / 2,
        )
        right_point = (
            (self.vertices[1][0] + self.vertices[2][0]) / 2,
            (self.vertices[1][1] + self.vertices[2][1]) / 2,
        )

        return [left_point, right_point]

    def get_direction_vector_alpha_distance(self) -> tuple[float, float]:
        left_point, right_point = self.get_direction_vector()
        alpha = (right_point[1] - left_point[1]) / (right_point[0] - left_point[0])
        distance = math.sqrt(
            (left_point[0] - right_point[0]) ** 2
            + (left_point[0] - right_point[0]) ** 2
        )
        return alpha, distance

    def detect_orientation(self) -> ImageOrientation:
        """Detect bounding poly orientation (up, down, left, or right).

        Google Vision vertices origin is at the top-left corner of the image.
        The order of each vertex gives the orientation of the text. For
        instance, horizontal text looks like this:
            0----1
            |    |
            3----2
        And left-rotated text looks like this:
            1----2
            |    |
            0----3

        See https://cloud.google.com/vision/docs/reference/rest/v1/images/annotate#Block
        for more details.

        We first select the two higher vertices of the image (lower y-values),
        and order the vertices by ascending x-value.

        This way, the index of these two vertices indicates the text
        orientation:
        - (0, 1) for horizontal (up, standard)
        - (1, 2) for 90° counterclockwise rotation (left)
        - (2, 3) for 180° rotation (down)
        - (3, 0) for 90° clockwise rotation (right)
        It is u
        """
        indexed_vertices = [(x[0], x[1], i) for i, x in enumerate(self.vertices)]
        # Sort by ascending y-value and select first two vertices:
        # get the two topmost vertices
        indexed_vertices = sorted(indexed_vertices, key=operator.itemgetter(1))[:2]

        first_vertex_index = indexed_vertices[0][2]
        second_vertex_index = indexed_vertices[1][2]

        # Sort by index ID, to make sure to filter permutations ((0, 1) and
        # not (1, 0))
        first_edge = tuple(sorted((first_vertex_index, second_vertex_index)))

        if first_edge == (0, 1):
            return ImageOrientation.up

        elif first_edge == (1, 2):
            return ImageOrientation.left

        elif first_edge == (2, 3):
            return ImageOrientation.down

        elif first_edge == (0, 3):
            return ImageOrientation.right

        else:
            logger.error(
                "Unknown orientation: edge %s, vertices %s",
                first_edge,
                self.vertices,
            )
            return ImageOrientation.unknown


class OCRTextAnnotation:
    __slots__ = ("locale", "text", "bounding_poly")

    def __init__(self, data: JSONType):
        self.locale = data.get("locale")
        self.text = data["description"]
        self.bounding_poly = BoundingPoly(data["boundingPoly"])


class LogoAnnotation:
    __slots__ = ("id", "description", "score")

    def __init__(self, data: JSONType):
        self.id = data.get("mid") or None
        self.score = data["score"]
        self.description = data["description"]


class LabelAnnotation:
    __slots__ = ("id", "description", "score")

    def __init__(self, data: JSONType):
        self.id = data.get("mid") or None
        self.score = data["score"]
        self.description = data["description"]


class SafeSearchAnnotation:
    __slots__ = ("adult", "spoof", "medical", "violence", "racy")

    def __init__(self, data: JSONType):
        self.adult: SafeSearchAnnotationLikelihood = SafeSearchAnnotationLikelihood[
            data["adult"]
        ]
        self.spoof: SafeSearchAnnotationLikelihood = SafeSearchAnnotationLikelihood[
            data["spoof"]
        ]
        self.medical: SafeSearchAnnotationLikelihood = SafeSearchAnnotationLikelihood[
            data["medical"]
        ]
        self.violence: SafeSearchAnnotationLikelihood = SafeSearchAnnotationLikelihood[
            data["violence"]
        ]
        self.racy: SafeSearchAnnotationLikelihood = SafeSearchAnnotationLikelihood[
            data["racy"]
        ]


class SafeSearchAnnotationLikelihood(enum.IntEnum):
    UNKNOWN = 1
    VERY_UNLIKELY = 2
    UNLIKELY = 3
    POSSIBLE = 4
    LIKELY = 5
    VERY_LIKELY = 6
